from typing import Dict, List, Optional

from attr import dataclass

from feast.feature import Feature
from feast.protos.feast.core.FeatureViewProjection_pb2 import (
    FeatureViewProjection as FeatureViewProjectionProto,
)


@dataclass
class FeatureViewProjection:
    name: str
    name_alias: Optional[str]
    features: List[Feature]
    join_key_map: Dict[str, str] = {}

    def name_to_use(self):
        return self.name_alias or self.name

    def to_proto(self):
        feature_reference_proto = FeatureViewProjectionProto(
            feature_view_name=self.name,
            feature_view_name_alias=self.name_alias,
            join_key_map=self.join_key_map,
        )
        for feature in self.features:
            feature_reference_proto.feature_columns.append(feature.to_proto())

        return feature_reference_proto

    @staticmethod
    def from_proto(proto: FeatureViewProjectionProto):
        ref = FeatureViewProjection(
            name=proto.feature_view_name,
            name_alias=proto.feature_view_name_alias,
            features=[],
            join_key_map=dict(proto.join_key_map),
        )
        for feature_column in proto.feature_columns:
            ref.features.append(Feature.from_proto(feature_column))

        return ref

    @staticmethod
    def from_definition(feature_grouping):
        return FeatureViewProjection(
            name=feature_grouping.name,
            name_alias=None,
            features=feature_grouping.features,
        )



from datetime import timedelta
from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd

from feast import Feature, FeatureView, OnDemandFeatureView, ValueType
from feast.data_source import DataSource, RequestDataSource
from feast.request_feature_view import RequestFeatureView


def driver_feature_view(
    data_source: DataSource,
    name="test_correctness",
    infer_features: bool = False,
    value_type: ValueType = ValueType.FLOAT,
) -> FeatureView:
    return FeatureView(
        name=name,
        entities=["driver"],
        features=None if infer_features else [Feature("value", value_type)],
        ttl=timedelta(days=5),
        input=data_source,
    )


def global_feature_view(
    data_source: DataSource,
    name="test_entityless",
    infer_features: bool = False,
    value_type: ValueType = ValueType.INT32,
) -> FeatureView:
    return FeatureView(
        name=name,
        entities=[],
        features=None if infer_features else [Feature("entityless_value", value_type)],
        ttl=timedelta(days=5),
        input=data_source,
    )


def conv_rate_plus_100(features_df: pd.DataFrame) -> pd.DataFrame:
    df = pd.DataFrame()
    df["conv_rate_plus_100"] = features_df["conv_rate"] + 100
    df["conv_rate_plus_val_to_add"] = (
        features_df["conv_rate"] + features_df["val_to_add"]
    )
    df["conv_rate_plus_100_rounded"] = (
        df["conv_rate_plus_100"].astype("float").round().astype(pd.Int32Dtype())
    )
    return df


def conv_rate_plus_100_feature_view(
    inputs: Dict[str, Union[RequestDataSource, FeatureView]],
    infer_features: bool = False,
    features: Optional[List[Feature]] = None,
) -> OnDemandFeatureView:
    _features = features or [
        Feature("conv_rate_plus_100", ValueType.DOUBLE),
        Feature("conv_rate_plus_val_to_add", ValueType.DOUBLE),
        Feature("conv_rate_plus_100_rounded", ValueType.INT32),
    ]
    return OnDemandFeatureView(
        name=conv_rate_plus_100.__name__,
        inputs=inputs,
        features=[] if infer_features else _features,
        udf=conv_rate_plus_100,
    )


def similarity(features_df: pd.DataFrame) -> pd.DataFrame:
    if features_df.size == 0:
        # give hint to Feast about return type
        df = pd.DataFrame({"cos_double": [0.0]})
        df["cos_float"] = df["cos_double"].astype(np.float32)
        return df
    vectors_a = features_df["embedding_double"].apply(np.array)
    vectors_b = features_df["vector_double"].apply(np.array)
    dot_products = vectors_a.mul(vectors_b).apply(sum)
    norms_q = vectors_a.apply(np.linalg.norm)
    norms_doc = vectors_b.apply(np.linalg.norm)
    df = pd.DataFrame()
    df["cos_double"] = dot_products / (norms_q * norms_doc)
    df["cos_float"] = df["cos_double"].astype(np.float32)
    return df


def similarity_feature_view(
    inputs: Dict[str, Union[RequestDataSource, FeatureView]],
    infer_features: bool = False,
    features: Optional[List[Feature]] = None,
) -> OnDemandFeatureView:
    _features = features or [
        Feature("cos_double", ValueType.DOUBLE),
        Feature("cos_float", ValueType.FLOAT),
    ]
    return OnDemandFeatureView(
        name=similarity.__name__,
        inputs=inputs,
        features=[] if infer_features else _features,
        udf=similarity,
    )


def create_driver_age_request_feature_view():
    return RequestFeatureView(
        name="driver_age",
        request_data_source=RequestDataSource(
            name="driver_age_source", schema={"driver_age": ValueType.INT32}
        ),
    )


def create_conv_rate_request_data_source():
    return RequestDataSource(
        name="conv_rate_input", schema={"val_to_add": ValueType.INT32}
    )


def create_similarity_request_data_source():
    return RequestDataSource(
        name="similarity_input",
        schema={
            "vector_double": ValueType.DOUBLE_LIST,
            "vector_float": ValueType.FLOAT_LIST,
        },
    )


def create_item_embeddings_feature_view(source, infer_features: bool = False):
    item_embeddings_feature_view = FeatureView(
        name="item_embeddings",
        entities=["item"],
        features=None
        if infer_features
        else [
            Feature(name="embedding_double", dtype=ValueType.DOUBLE_LIST),
            Feature(name="embedding_float", dtype=ValueType.FLOAT_LIST),
        ],
        batch_source=source,
        ttl=timedelta(hours=2),
    )
    return item_embeddings_feature_view


def create_driver_hourly_stats_feature_view(source, infer_features: bool = False):
    driver_stats_feature_view = FeatureView(
        name="driver_stats",
        entities=["driver"],
        features=None
        if infer_features
        else [
            Feature(name="conv_rate", dtype=ValueType.FLOAT),
            Feature(name="acc_rate", dtype=ValueType.FLOAT),
            Feature(name="avg_daily_trips", dtype=ValueType.INT32),
        ],
        batch_source=source,
        ttl=timedelta(hours=2),
    )
    return driver_stats_feature_view


def create_customer_daily_profile_feature_view(source, infer_features: bool = False):
    customer_profile_feature_view = FeatureView(
        name="customer_profile",
        entities=["customer_id"],
        features=None
        if infer_features
        else [
            Feature(name="current_balance", dtype=ValueType.FLOAT),
            Feature(name="avg_passenger_count", dtype=ValueType.FLOAT),
            Feature(name="lifetime_trip_count", dtype=ValueType.INT32),
        ],
        batch_source=source,
        ttl=timedelta(days=2),
    )
    return customer_profile_feature_view


def create_global_stats_feature_view(source, infer_features: bool = False):
    global_stats_feature_view = FeatureView(
        name="global_stats",
        entities=[],
        features=None
        if infer_features
        else [
            Feature(name="num_rides", dtype=ValueType.INT32),
            Feature(name="avg_ride_length", dtype=ValueType.FLOAT),
        ],
        batch_source=source,
        ttl=timedelta(days=2),
    )
    return global_stats_feature_view


def create_order_feature_view(source, infer_features: bool = False):
    return FeatureView(
        name="order",
        entities=["driver", "customer_id"],
        features=None
        if infer_features
        else [Feature(name="order_is_success", dtype=ValueType.INT32)],
        batch_source=source,
        ttl=timedelta(days=2),
    )


def create_location_stats_feature_view(source, infer_features: bool = False):
    location_stats_feature_view = FeatureView(
        name="location_stats",
        entities=["location_id"],
        features=None
        if infer_features
        else [Feature(name="temperature", dtype=ValueType.INT32)],
        batch_source=source,
        ttl=timedelta(days=2),
    )
    return location_stats_feature_view



import datetime
import itertools
import os
import time
import unittest
from datetime import timedelta
from typing import Any, Dict, List, Union

import assertpy
import numpy as np
import pandas as pd
import pytest
import requests

from feast import Entity, Feature, FeatureService, FeatureView, ValueType
from feast.errors import (
    FeatureNameCollisionError,
    RequestDataNotFoundInEntityRowsException,
)
from tests.integration.feature_repos.repo_configuration import (
    Environment,
    construct_universal_feature_views,
)
from tests.integration.feature_repos.universal.entities import (
    customer,
    driver,
    location,
)
from tests.integration.feature_repos.universal.feature_views import (
    create_driver_hourly_stats_feature_view,
)
from tests.utils.data_source_utils import prep_file_source


# TODO: make this work with all universal (all online store types)
@pytest.mark.integration
def test_write_to_online_store_event_check(local_redis_environment):
    if os.getenv("FEAST_IS_LOCAL_TEST", "False") == "True":
        return
    fs = local_redis_environment.feature_store

    # write same data points 3 with different timestamps
    now = pd.Timestamp(datetime.datetime.utcnow()).round("ms")
    hour_ago = pd.Timestamp(datetime.datetime.utcnow() - timedelta(hours=1)).round("ms")
    latest = pd.Timestamp(datetime.datetime.utcnow() + timedelta(seconds=1)).round("ms")

    data = {
        "id": [123, 567, 890],
        "string_col": ["OLD_FEATURE", "LATEST_VALUE2", "LATEST_VALUE3"],
        "ts_1": [hour_ago, now, now],
    }
    dataframe_source = pd.DataFrame(data)
    with prep_file_source(
        df=dataframe_source, event_timestamp_column="ts_1"
    ) as file_source:
        e = Entity(name="id", value_type=ValueType.STRING)

        # Create Feature View
        fv1 = FeatureView(
            name="feature_view_123",
            features=[Feature(name="string_col", dtype=ValueType.STRING)],
            entities=["id"],
            batch_source=file_source,
            ttl=timedelta(minutes=5),
        )
        # Register Feature View and Entity
        fs.apply([fv1, e])

        #  data to ingest into Online Store (recent)
        data = {
            "id": [123],
            "string_col": ["hi_123"],
            "ts_1": [now],
        }
        df_data = pd.DataFrame(data)

        # directly ingest data into the Online Store
        fs.write_to_online_store("feature_view_123", df_data)

        df = fs.get_online_features(
            features=["feature_view_123:string_col"], entity_rows=[{"id": 123}]
        ).to_df()
        assert df["string_col"].iloc[0] == "hi_123"

        # data to ingest into Online Store (1 hour delayed data)
        # should now overwrite features for id=123 because it's less recent data
        data = {
            "id": [123, 567, 890],
            "string_col": ["bye_321", "hello_123", "greetings_321"],
            "ts_1": [hour_ago, hour_ago, hour_ago],
        }
        df_data = pd.DataFrame(data)

        # directly ingest data into the Online Store
        fs.write_to_online_store("feature_view_123", df_data)

        df = fs.get_online_features(
            features=["feature_view_123:string_col"],
            entity_rows=[{"id": 123}, {"id": 567}, {"id": 890}],
        ).to_df()
        assert df["string_col"].iloc[0] == "hi_123"
        assert df["string_col"].iloc[1] == "hello_123"
        assert df["string_col"].iloc[2] == "greetings_321"

        # should overwrite string_col for id=123 because it's most recent based on event_timestamp
        data = {
            "id": [123],
            "string_col": ["LATEST_VALUE"],
            "ts_1": [latest],
        }
        df_data = pd.DataFrame(data)

        fs.write_to_online_store("feature_view_123", df_data)

        df = fs.get_online_features(
            features=["feature_view_123:string_col"],
            entity_rows=[{"id": 123}, {"id": 567}, {"id": 890}],
        ).to_df()
        assert df["string_col"].iloc[0] == "LATEST_VALUE"
        assert df["string_col"].iloc[1] == "hello_123"
        assert df["string_col"].iloc[2] == "greetings_321"

        # writes to online store via datasource (dataframe_source) materialization
        fs.materialize(
            start_date=datetime.datetime.now() - timedelta(hours=12),
            end_date=datetime.datetime.utcnow(),
        )

        df = fs.get_online_features(
            features=["feature_view_123:string_col"],
            entity_rows=[{"id": 123}, {"id": 567}, {"id": 890}],
        ).to_df()
        assert df["string_col"].iloc[0] == "LATEST_VALUE"
        assert df["string_col"].iloc[1] == "LATEST_VALUE2"
        assert df["string_col"].iloc[2] == "LATEST_VALUE3"


@pytest.mark.integration
def test_write_to_online_store(environment, universal_data_sources):
    fs = environment.feature_store
    entities, datasets, data_sources = universal_data_sources
    driver_hourly_stats = create_driver_hourly_stats_feature_view(
        data_sources["driver"]
    )
    driver_entity = driver()

    # Register Feature View and Entity
    fs.apply([driver_hourly_stats, driver_entity])

    # fake data to ingest into Online Store
    data = {
        "driver_id": [123],
        "conv_rate": [0.85],
        "acc_rate": [0.91],
        "avg_daily_trips": [14],
        "event_timestamp": [pd.Timestamp(datetime.datetime.utcnow()).round("ms")],
        "created": [pd.Timestamp(datetime.datetime.utcnow()).round("ms")],
    }
    df_data = pd.DataFrame(data)

    # directly ingest data into the Online Store
    fs.write_to_online_store("driver_stats", df_data)

    # assert the right data is in the Online Store
    df = fs.get_online_features(
        features=[
            "driver_stats:avg_daily_trips",
            "driver_stats:acc_rate",
            "driver_stats:conv_rate",
        ],
        entity_rows=[{"driver": 123}],
    ).to_df()
    assertpy.assert_that(df["avg_daily_trips"].iloc[0]).is_equal_to(14)
    assertpy.assert_that(df["acc_rate"].iloc[0]).is_close_to(0.91, 1e-6)
    assertpy.assert_that(df["conv_rate"].iloc[0]).is_close_to(0.85, 1e-6)


def _get_online_features_dict_remotely(
    endpoint: str,
    features: Union[List[str], FeatureService],
    entity_rows: List[Dict[str, Any]],
    full_feature_names: bool = False,
) -> Dict[str, List[Any]]:
    """Sends the online feature request to a remote feature server (through endpoint) and returns the feature dict.

    The output should be identical to:

    >>> fs.get_online_features(features=features, entity_rows=entity_rows, full_feature_names=full_feature_names).to_dict()

    This makes it easy to test the remote feature server by comparing the output to the local method.

    """
    request = {
        # Convert list of dicts (entity_rows) into dict of lists (entities) for json request
        "entities": {key: [row[key] for row in entity_rows] for key in entity_rows[0]},
        "full_feature_names": full_feature_names,
    }
    # Either set features of feature_service depending on the parameter
    if isinstance(features, list):
        request["features"] = features
    else:
        request["feature_service"] = features.name
    for _ in range(25):
        # Send the request to the remote feature server and get the response in JSON format
        response = requests.post(
            f"{endpoint}/get-online-features", json=request, timeout=30
        ).json()
        # Retry if the response is internal server error, which can happen when lambda is being restarted
        if response.get("message") != "Internal Server Error":
            break
        # Sleep between retries to give the server some time to start
        time.sleep(1)
    else:
        raise Exception("Failed to get online features from remote feature server")
    keys = response["field_values"][0]["statuses"].keys()
    # Get rid of unnecessary structure in the response, leaving list of dicts
    response = [row["fields"] for row in response["field_values"]]
    # Convert list of dicts (response) into dict of lists which is the format of the return value
    return {key: [row.get(key) for row in response] for key in keys}


def get_online_features_dict(
    environment: Environment,
    features: Union[List[str], FeatureService],
    entity_rows: List[Dict[str, Any]],
    full_feature_names: bool = False,
) -> Dict[str, List[Any]]:
    """Get the online feature values from both SDK and remote feature servers, assert equality and return values.

    Always use this method instead of fs.get_online_features(...) in this test file.

    """
    online_features = environment.feature_store.get_online_features(
        features=features,
        entity_rows=entity_rows,
        full_feature_names=full_feature_names,
    )
    assertpy.assert_that(online_features).is_not_none()
    dict1 = online_features.to_dict()

    endpoint = environment.feature_store.get_feature_server_endpoint()
    # If endpoint is None, it means that the remote feature server isn't configured
    if endpoint is not None:
        dict2 = _get_online_features_dict_remotely(
            endpoint=endpoint,
            features=features,
            entity_rows=entity_rows,
            full_feature_names=full_feature_names,
        )

        # Make sure that the two dicts are equal
        assertpy.assert_that(dict1).is_equal_to(dict2)
    elif environment.python_feature_server:
        raise ValueError(
            "feature_store.get_feature_server_endpoint() is None while python feature server is enabled"
        )
    return dict1


@pytest.mark.integration
@pytest.mark.universal
@pytest.mark.parametrize("full_feature_names", [True, False], ids=lambda v: str(v))
def test_online_retrieval(environment, universal_data_sources, full_feature_names):
    fs = environment.feature_store
    entities, datasets, data_sources = universal_data_sources
    feature_views = construct_universal_feature_views(data_sources)

    feature_service = FeatureService(
        "convrate_plus100",
        features=[
            feature_views["driver"][["conv_rate"]],
            feature_views["driver_odfv"],
            feature_views["driver_age_request_fv"],
        ],
    )
    feature_service_entity_mapping = FeatureService(
        name="entity_mapping",
        features=[
            feature_views["location"]
            .with_name("origin")
            .with_join_key_map({"location_id": "origin_id"}),
            feature_views["location"]
            .with_name("destination")
            .with_join_key_map({"location_id": "destination_id"}),
        ],
    )

    feast_objects = []
    feast_objects.extend(feature_views.values())
    feast_objects.extend(
        [
            driver(),
            customer(),
            location(),
            feature_service,
            feature_service_entity_mapping,
        ]
    )
    fs.apply(feast_objects)
    fs.materialize(
        environment.start_date - timedelta(days=1),
        environment.end_date + timedelta(days=1),
    )

    entity_sample = datasets["orders"].sample(10)[
        ["customer_id", "driver_id", "order_id", "event_timestamp"]
    ]
    orders_df = datasets["orders"][
        (
            datasets["orders"]["customer_id"].isin(entity_sample["customer_id"])
            & datasets["orders"]["driver_id"].isin(entity_sample["driver_id"])
        )
    ]

    sample_drivers = entity_sample["driver_id"]
    drivers_df = datasets["driver"][
        datasets["driver"]["driver_id"].isin(sample_drivers)
    ]

    sample_customers = entity_sample["customer_id"]
    customers_df = datasets["customer"][
        datasets["customer"]["customer_id"].isin(sample_customers)
    ]

    location_pairs = np.array(list(itertools.permutations(entities["location"], 2)))
    sample_location_pairs = location_pairs[
        np.random.choice(len(location_pairs), 10)
    ].T.tolist()
    origins_df = datasets["location"][
        datasets["location"]["location_id"].isin(sample_location_pairs[0])
    ]
    destinations_df = datasets["location"][
        datasets["location"]["location_id"].isin(sample_location_pairs[1])
    ]

    global_df = datasets["global"]

    entity_rows = [
        {"driver": d, "customer_id": c, "val_to_add": 50, "driver_age": 25}
        for (d, c) in zip(sample_drivers, sample_customers)
    ]

    feature_refs = [
        "driver_stats:conv_rate",
        "driver_stats:avg_daily_trips",
        "customer_profile:current_balance",
        "customer_profile:avg_passenger_count",
        "customer_profile:lifetime_trip_count",
        "conv_rate_plus_100:conv_rate_plus_100",
        "conv_rate_plus_100:conv_rate_plus_val_to_add",
        "order:order_is_success",
        "global_stats:num_rides",
        "global_stats:avg_ride_length",
        "driver_age:driver_age",
    ]
    unprefixed_feature_refs = [f.rsplit(":", 1)[-1] for f in feature_refs if ":" in f]
    # Remove the on demand feature view output features, since they're not present in the source dataframe
    unprefixed_feature_refs.remove("conv_rate_plus_100")
    unprefixed_feature_refs.remove("conv_rate_plus_val_to_add")

    online_features_dict = get_online_features_dict(
        environment=environment,
        features=feature_refs,
        entity_rows=entity_rows,
        full_feature_names=full_feature_names,
    )

    # Test that the on demand feature views compute properly even if the dependent conv_rate
    # feature isn't requested.
    online_features_no_conv_rate = get_online_features_dict(
        environment=environment,
        features=[ref for ref in feature_refs if ref != "driver_stats:conv_rate"],
        entity_rows=entity_rows,
        full_feature_names=full_feature_names,
    )

    assert online_features_no_conv_rate is not None

    keys = online_features_dict.keys()
    assert (
        len(keys) == len(feature_refs) + 2
    )  # Add two for the driver id and the customer id entity keys
    for feature in feature_refs:
        # full_feature_names does not apply to request feature views
        if full_feature_names and feature != "driver_age:driver_age":
            assert feature.replace(":", "__") in keys
        else:
            assert feature.rsplit(":", 1)[-1] in keys
            assert (
                "driver_stats" not in keys
                and "customer_profile" not in keys
                and "order" not in keys
                and "global_stats" not in keys
            )

    tc = unittest.TestCase()
    for i, entity_row in enumerate(entity_rows):
        df_features = get_latest_feature_values_from_dataframes(
            driver_df=drivers_df,
            customer_df=customers_df,
            orders_df=orders_df,
            global_df=global_df,
            entity_row=entity_row,
        )

        assert df_features["customer_id"] == online_features_dict["customer_id"][i]
        assert df_features["driver_id"] == online_features_dict["driver_id"][i]
        tc.assertAlmostEqual(
            online_features_dict[
                response_feature_name("conv_rate_plus_100", full_feature_names)
            ][i],
            df_features["conv_rate"] + 100,
            delta=0.0001,
        )
        tc.assertAlmostEqual(
            online_features_dict[
                response_feature_name("conv_rate_plus_val_to_add", full_feature_names)
            ][i],
            df_features["conv_rate"] + df_features["val_to_add"],
            delta=0.0001,
        )
        for unprefixed_feature_ref in unprefixed_feature_refs:
            tc.assertAlmostEqual(
                df_features[unprefixed_feature_ref],
                online_features_dict[
                    response_feature_name(unprefixed_feature_ref, full_feature_names)
                ][i],
                delta=0.0001,
            )

    # Check what happens for missing values
    missing_responses_dict = get_online_features_dict(
        environment=environment,
        features=feature_refs,
        entity_rows=[
            {"driver": 0, "customer_id": 0, "val_to_add": 100, "driver_age": 125}
        ],
        full_feature_names=full_feature_names,
    )
    assert missing_responses_dict is not None
    for unprefixed_feature_ref in unprefixed_feature_refs:
        if unprefixed_feature_ref not in {"num_rides", "avg_ride_length", "driver_age"}:
            tc.assertIsNone(
                missing_responses_dict[
                    response_feature_name(unprefixed_feature_ref, full_feature_names)
                ][0]
            )

    # Check what happens for missing request data
    with pytest.raises(RequestDataNotFoundInEntityRowsException):
        get_online_features_dict(
            environment=environment,
            features=feature_refs,
            entity_rows=[{"driver": 0, "customer_id": 0}],
            full_feature_names=full_feature_names,
        )

    # Also with request data
    with pytest.raises(RequestDataNotFoundInEntityRowsException):
        get_online_features_dict(
            environment=environment,
            features=feature_refs,
            entity_rows=[{"driver": 0, "customer_id": 0, "val_to_add": 20}],
            full_feature_names=full_feature_names,
        )

    assert_feature_service_correctness(
        environment,
        feature_service,
        entity_rows,
        full_feature_names,
        drivers_df,
        customers_df,
        orders_df,
        global_df,
    )

    entity_rows = [
        {
            "driver": driver,
            "customer_id": customer,
            "origin_id": origin,
            "destination_id": destination,
        }
        for (driver, customer, origin, destination) in zip(
            sample_drivers, sample_customers, *sample_location_pairs
        )
    ]
    assert_feature_service_entity_mapping_correctness(
        environment,
        feature_service_entity_mapping,
        entity_rows,
        full_feature_names,
        drivers_df,
        customers_df,
        orders_df,
        origins_df,
        destinations_df,
    )


def response_feature_name(feature: str, full_feature_names: bool) -> str:
    if (
        feature in {"current_balance", "avg_passenger_count", "lifetime_trip_count"}
        and full_feature_names
    ):
        return f"customer_profile__{feature}"

    if feature in {"conv_rate", "avg_daily_trips"} and full_feature_names:
        return f"driver_stats__{feature}"

    if (
        feature in {"conv_rate_plus_100", "conv_rate_plus_val_to_add"}
        and full_feature_names
    ):
        return f"conv_rate_plus_100__{feature}"

    if feature in {"order_is_success"} and full_feature_names:
        return f"order__{feature}"

    if feature in {"num_rides", "avg_ride_length"} and full_feature_names:
        return f"global_stats__{feature}"

    return feature


def get_latest_row(entity_row, df, join_key, entity_key):
    rows = df[df[join_key] == entity_row[entity_key]]
    return rows.loc[rows["event_timestamp"].idxmax()].to_dict()


def get_latest_feature_values_from_dataframes(
    driver_df,
    customer_df,
    orders_df,
    entity_row,
    global_df=None,
    origin_df=None,
    destination_df=None,
):
    latest_driver_row = get_latest_row(entity_row, driver_df, "driver_id", "driver")
    latest_customer_row = get_latest_row(
        entity_row, customer_df, "customer_id", "customer_id"
    )

    # Since the event timestamp columns may contain timestamps of different timezones,
    # we must first convert the timestamps to UTC before we can compare them.
    order_rows = orders_df[
        (orders_df["driver_id"] == entity_row["driver"])
        & (orders_df["customer_id"] == entity_row["customer_id"])
    ]
    timestamps = order_rows[["event_timestamp"]]
    timestamps["event_timestamp"] = pd.to_datetime(
        timestamps["event_timestamp"], utc=True
    )
    max_index = timestamps["event_timestamp"].idxmax()
    latest_orders_row = order_rows.loc[max_index]

    if global_df is not None:
        latest_global_row = global_df.loc[
            global_df["event_timestamp"].idxmax()
        ].to_dict()
    if origin_df is not None:
        latest_origin_row = get_latest_row(
            entity_row, origin_df, "location_id", "origin_id"
        )
        latest_destination_row = get_latest_row(
            entity_row, destination_df, "location_id", "destination_id"
        )
        # Need full feature names for shadow entities
        latest_origin_row["origin__temperature"] = latest_origin_row.pop("temperature")
        latest_destination_row["destination__temperature"] = latest_destination_row.pop(
            "temperature"
        )
    request_data_features = entity_row.copy()
    request_data_features.pop("driver")
    request_data_features.pop("customer_id")
    if global_df is not None:
        return {
            **latest_customer_row,
            **latest_driver_row,
            **latest_orders_row,
            **latest_global_row,
            **request_data_features,
        }
    if origin_df is not None:
        request_data_features.pop("origin_id")
        request_data_features.pop("destination_id")
        return {
            **latest_customer_row,
            **latest_driver_row,
            **latest_orders_row,
            **latest_origin_row,
            **latest_destination_row,
            **request_data_features,
        }
    return {
        **latest_customer_row,
        **latest_driver_row,
        **latest_orders_row,
        **request_data_features,
    }


def assert_feature_service_correctness(
    environment,
    feature_service,
    entity_rows,
    full_feature_names,
    drivers_df,
    customers_df,
    orders_df,
    global_df,
):
    feature_service_online_features_dict = get_online_features_dict(
        environment=environment,
        features=feature_service,
        entity_rows=entity_rows,
        full_feature_names=full_feature_names,
    )
    feature_service_keys = feature_service_online_features_dict.keys()

    assert (
        len(feature_service_keys)
        == sum(
            [
                len(projection.features)
                for projection in feature_service.feature_view_projections
            ]
        )
        + 2
    )  # Add two for the driver id and the customer id entity keys

    tc = unittest.TestCase()
    for i, entity_row in enumerate(entity_rows):
        df_features = get_latest_feature_values_from_dataframes(
            driver_df=drivers_df,
            customer_df=customers_df,
            orders_df=orders_df,
            global_df=global_df,
            entity_row=entity_row,
        )
        tc.assertAlmostEqual(
            feature_service_online_features_dict[
                response_feature_name("conv_rate_plus_100", full_feature_names)
            ][i],
            df_features["conv_rate"] + 100,
            delta=0.0001,
        )


def assert_feature_service_entity_mapping_correctness(
    environment,
    feature_service,
    entity_rows,
    full_feature_names,
    drivers_df,
    customers_df,
    orders_df,
    origins_df,
    destinations_df,
):
    if full_feature_names:
        feature_service_online_features_dict = get_online_features_dict(
            environment=environment,
            features=feature_service,
            entity_rows=entity_rows,
            full_feature_names=full_feature_names,
        )
        feature_service_keys = feature_service_online_features_dict.keys()

        assert (
            len(feature_service_keys)
            == sum(
                [
                    len(projection.features)
                    for projection in feature_service.feature_view_projections
                ]
            )
            + 4
        )  # Add 4 for the driver_id, customer_id, origin_id, and destination_id entity keys

        for i, entity_row in enumerate(entity_rows):
            df_features = get_latest_feature_values_from_dataframes(
                driver_df=drivers_df,
                customer_df=customers_df,
                orders_df=orders_df,
                origin_df=origins_df,
                destination_df=destinations_df,
                entity_row=entity_row,
            )
            for feature_name in ["origin__temperature", "destination__temperature"]:
                assert (
                    feature_service_online_features_dict[feature_name][i]
                    == df_features[feature_name]
                )
    else:
        # using 2 of the same FeatureView without full_feature_names=True will result in collision
        with pytest.raises(FeatureNameCollisionError):
            get_online_features_dict(
                environment=environment,
                features=feature_service,
                entity_rows=entity_rows,
                full_feature_names=full_feature_names,
            )



FLAG_ALPHA_FEATURES_NAME = "alpha_features"
FLAG_ON_DEMAND_TRANSFORM_NAME = "on_demand_transforms"
FLAG_PYTHON_FEATURE_SERVER_NAME = "python_feature_server"
FLAG_AWS_LAMBDA_FEATURE_SERVER_NAME = "aws_lambda_feature_server"
FLAG_DIRECT_INGEST_TO_ONLINE_STORE = "direct_ingest_to_online_store"
ENV_FLAG_IS_TEST = "IS_TEST"

FLAG_NAMES = {
    FLAG_ALPHA_FEATURES_NAME,
    FLAG_ON_DEMAND_TRANSFORM_NAME,
    FLAG_PYTHON_FEATURE_SERVER_NAME,
    FLAG_AWS_LAMBDA_FEATURE_SERVER_NAME,
    FLAG_DIRECT_INGEST_TO_ONLINE_STORE,
}



import importlib

from feast import errors


def get_class_from_type(module_name: str, class_name: str, class_type: str):
    if not class_name.endswith(class_type):
        raise errors.FeastClassInvalidName(class_name, class_type)

    # Try importing the module that contains the custom provider
    try:
        module = importlib.import_module(module_name)
    except Exception as e:
        # The original exception can be anything - either module not found,
        # or any other kind of error happening during the module import time.
        # So we should include the original error as well in the stack trace.
        raise errors.FeastModuleImportError(module_name, class_type) from e

    # Try getting the provider class definition
    try:
        _class = getattr(module, class_name)
    except AttributeError:
        # This can only be one type of error, when class_name attribute does not exist in the module
        # So we don't have to include the original exception here
        raise errors.FeastClassImportError(
            module_name, class_name, class_type=class_type
        ) from None
    return _class



from feast import FeatureService


def test_feature_service_with_description():
    feature_service = FeatureService(
        name="my-feature-service", features=[], description="a clear description"
    )
    assert feature_service.to_proto().spec.description == "a clear description"


def test_feature_service_without_description():
    feature_service = FeatureService(name="my-feature-service", features=[])
    #
    assert feature_service.to_proto().spec.description == ""



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
from google.protobuf.any_pb2 import (
    Any as google___protobuf___any_pb2___Any,
)

from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
    EnumDescriptor as google___protobuf___descriptor___EnumDescriptor,
    FileDescriptor as google___protobuf___descriptor___FileDescriptor,
)

from google.protobuf.internal.containers import (
    MessageMap as google___protobuf___internal___containers___MessageMap,
    RepeatedCompositeFieldContainer as google___protobuf___internal___containers___RepeatedCompositeFieldContainer,
    RepeatedScalarFieldContainer as google___protobuf___internal___containers___RepeatedScalarFieldContainer,
)

from google.protobuf.internal.enum_type_wrapper import (
    _EnumTypeWrapper as google___protobuf___internal___enum_type_wrapper____EnumTypeWrapper,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from tensorflow_metadata.proto.v0.path_pb2 import (
    Path as tensorflow_metadata___proto___v0___path_pb2___Path,
)

from typing import (
    Iterable as typing___Iterable,
    Mapping as typing___Mapping,
    NewType as typing___NewType,
    Optional as typing___Optional,
    Text as typing___Text,
    cast as typing___cast,
    overload as typing___overload,
)

from typing_extensions import (
    Literal as typing_extensions___Literal,
)


builtin___bool = bool
builtin___bytes = bytes
builtin___float = float
builtin___int = int


DESCRIPTOR: google___protobuf___descriptor___FileDescriptor = ...

LifecycleStageValue = typing___NewType('LifecycleStageValue', builtin___int)
type___LifecycleStageValue = LifecycleStageValue
LifecycleStage: _LifecycleStage
class _LifecycleStage(google___protobuf___internal___enum_type_wrapper____EnumTypeWrapper[LifecycleStageValue]):
    DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
    UNKNOWN_STAGE = typing___cast(LifecycleStageValue, 0)
    PLANNED = typing___cast(LifecycleStageValue, 1)
    ALPHA = typing___cast(LifecycleStageValue, 2)
    BETA = typing___cast(LifecycleStageValue, 3)
    PRODUCTION = typing___cast(LifecycleStageValue, 4)
    DEPRECATED = typing___cast(LifecycleStageValue, 5)
    DEBUG_ONLY = typing___cast(LifecycleStageValue, 6)
UNKNOWN_STAGE = typing___cast(LifecycleStageValue, 0)
PLANNED = typing___cast(LifecycleStageValue, 1)
ALPHA = typing___cast(LifecycleStageValue, 2)
BETA = typing___cast(LifecycleStageValue, 3)
PRODUCTION = typing___cast(LifecycleStageValue, 4)
DEPRECATED = typing___cast(LifecycleStageValue, 5)
DEBUG_ONLY = typing___cast(LifecycleStageValue, 6)

FeatureTypeValue = typing___NewType('FeatureTypeValue', builtin___int)
type___FeatureTypeValue = FeatureTypeValue
FeatureType: _FeatureType
class _FeatureType(google___protobuf___internal___enum_type_wrapper____EnumTypeWrapper[FeatureTypeValue]):
    DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
    TYPE_UNKNOWN = typing___cast(FeatureTypeValue, 0)
    BYTES = typing___cast(FeatureTypeValue, 1)
    INT = typing___cast(FeatureTypeValue, 2)
    FLOAT = typing___cast(FeatureTypeValue, 3)
    STRUCT = typing___cast(FeatureTypeValue, 4)
TYPE_UNKNOWN = typing___cast(FeatureTypeValue, 0)
BYTES = typing___cast(FeatureTypeValue, 1)
INT = typing___cast(FeatureTypeValue, 2)
FLOAT = typing___cast(FeatureTypeValue, 3)
STRUCT = typing___cast(FeatureTypeValue, 4)

class Schema(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class TensorRepresentationGroupEntry(google___protobuf___message___Message):
        DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
        key: typing___Text = ...

        @property
        def value(self) -> type___TensorRepresentationGroup: ...

        def __init__(self,
            *,
            key : typing___Optional[typing___Text] = None,
            value : typing___Optional[type___TensorRepresentationGroup] = None,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions___Literal[u"key",b"key",u"value",b"value"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"key",b"key",u"value",b"value"]) -> None: ...
    type___TensorRepresentationGroupEntry = TensorRepresentationGroupEntry

    default_environment: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text] = ...

    @property
    def feature(self) -> google___protobuf___internal___containers___RepeatedCompositeFieldContainer[type___Feature]: ...

    @property
    def sparse_feature(self) -> google___protobuf___internal___containers___RepeatedCompositeFieldContainer[type___SparseFeature]: ...

    @property
    def weighted_feature(self) -> google___protobuf___internal___containers___RepeatedCompositeFieldContainer[type___WeightedFeature]: ...

    @property
    def string_domain(self) -> google___protobuf___internal___containers___RepeatedCompositeFieldContainer[type___StringDomain]: ...

    @property
    def float_domain(self) -> google___protobuf___internal___containers___RepeatedCompositeFieldContainer[type___FloatDomain]: ...

    @property
    def int_domain(self) -> google___protobuf___internal___containers___RepeatedCompositeFieldContainer[type___IntDomain]: ...

    @property
    def annotation(self) -> type___Annotation: ...

    @property
    def dataset_constraints(self) -> type___DatasetConstraints: ...

    @property
    def tensor_representation_group(self) -> google___protobuf___internal___containers___MessageMap[typing___Text, type___TensorRepresentationGroup]: ...

    def __init__(self,
        *,
        feature : typing___Optional[typing___Iterable[type___Feature]] = None,
        sparse_feature : typing___Optional[typing___Iterable[type___SparseFeature]] = None,
        weighted_feature : typing___Optional[typing___Iterable[type___WeightedFeature]] = None,
        string_domain : typing___Optional[typing___Iterable[type___StringDomain]] = None,
        float_domain : typing___Optional[typing___Iterable[type___FloatDomain]] = None,
        int_domain : typing___Optional[typing___Iterable[type___IntDomain]] = None,
        default_environment : typing___Optional[typing___Iterable[typing___Text]] = None,
        annotation : typing___Optional[type___Annotation] = None,
        dataset_constraints : typing___Optional[type___DatasetConstraints] = None,
        tensor_representation_group : typing___Optional[typing___Mapping[typing___Text, type___TensorRepresentationGroup]] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"annotation",b"annotation",u"dataset_constraints",b"dataset_constraints"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"annotation",b"annotation",u"dataset_constraints",b"dataset_constraints",u"default_environment",b"default_environment",u"feature",b"feature",u"float_domain",b"float_domain",u"int_domain",b"int_domain",u"sparse_feature",b"sparse_feature",u"string_domain",b"string_domain",u"tensor_representation_group",b"tensor_representation_group",u"weighted_feature",b"weighted_feature"]) -> None: ...
type___Schema = Schema

class Feature(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    name: typing___Text = ...
    deprecated: builtin___bool = ...
    type: type___FeatureTypeValue = ...
    domain: typing___Text = ...
    in_environment: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text] = ...
    not_in_environment: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text] = ...
    lifecycle_stage: type___LifecycleStageValue = ...

    @property
    def presence(self) -> type___FeaturePresence: ...

    @property
    def group_presence(self) -> type___FeaturePresenceWithinGroup: ...

    @property
    def shape(self) -> type___FixedShape: ...

    @property
    def value_count(self) -> type___ValueCount: ...

    @property
    def int_domain(self) -> type___IntDomain: ...

    @property
    def float_domain(self) -> type___FloatDomain: ...

    @property
    def string_domain(self) -> type___StringDomain: ...

    @property
    def bool_domain(self) -> type___BoolDomain: ...

    @property
    def struct_domain(self) -> type___StructDomain: ...

    @property
    def natural_language_domain(self) -> type___NaturalLanguageDomain: ...

    @property
    def image_domain(self) -> type___ImageDomain: ...

    @property
    def mid_domain(self) -> type___MIDDomain: ...

    @property
    def url_domain(self) -> type___URLDomain: ...

    @property
    def time_domain(self) -> type___TimeDomain: ...

    @property
    def time_of_day_domain(self) -> type___TimeOfDayDomain: ...

    @property
    def distribution_constraints(self) -> type___DistributionConstraints: ...

    @property
    def annotation(self) -> type___Annotation: ...

    @property
    def skew_comparator(self) -> type___FeatureComparator: ...

    @property
    def drift_comparator(self) -> type___FeatureComparator: ...

    def __init__(self,
        *,
        name : typing___Optional[typing___Text] = None,
        deprecated : typing___Optional[builtin___bool] = None,
        presence : typing___Optional[type___FeaturePresence] = None,
        group_presence : typing___Optional[type___FeaturePresenceWithinGroup] = None,
        shape : typing___Optional[type___FixedShape] = None,
        value_count : typing___Optional[type___ValueCount] = None,
        type : typing___Optional[type___FeatureTypeValue] = None,
        domain : typing___Optional[typing___Text] = None,
        int_domain : typing___Optional[type___IntDomain] = None,
        float_domain : typing___Optional[type___FloatDomain] = None,
        string_domain : typing___Optional[type___StringDomain] = None,
        bool_domain : typing___Optional[type___BoolDomain] = None,
        struct_domain : typing___Optional[type___StructDomain] = None,
        natural_language_domain : typing___Optional[type___NaturalLanguageDomain] = None,
        image_domain : typing___Optional[type___ImageDomain] = None,
        mid_domain : typing___Optional[type___MIDDomain] = None,
        url_domain : typing___Optional[type___URLDomain] = None,
        time_domain : typing___Optional[type___TimeDomain] = None,
        time_of_day_domain : typing___Optional[type___TimeOfDayDomain] = None,
        distribution_constraints : typing___Optional[type___DistributionConstraints] = None,
        annotation : typing___Optional[type___Annotation] = None,
        skew_comparator : typing___Optional[type___FeatureComparator] = None,
        drift_comparator : typing___Optional[type___FeatureComparator] = None,
        in_environment : typing___Optional[typing___Iterable[typing___Text]] = None,
        not_in_environment : typing___Optional[typing___Iterable[typing___Text]] = None,
        lifecycle_stage : typing___Optional[type___LifecycleStageValue] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"annotation",b"annotation",u"bool_domain",b"bool_domain",u"deprecated",b"deprecated",u"distribution_constraints",b"distribution_constraints",u"domain",b"domain",u"domain_info",b"domain_info",u"drift_comparator",b"drift_comparator",u"float_domain",b"float_domain",u"group_presence",b"group_presence",u"image_domain",b"image_domain",u"int_domain",b"int_domain",u"lifecycle_stage",b"lifecycle_stage",u"mid_domain",b"mid_domain",u"name",b"name",u"natural_language_domain",b"natural_language_domain",u"presence",b"presence",u"presence_constraints",b"presence_constraints",u"shape",b"shape",u"shape_type",b"shape_type",u"skew_comparator",b"skew_comparator",u"string_domain",b"string_domain",u"struct_domain",b"struct_domain",u"time_domain",b"time_domain",u"time_of_day_domain",b"time_of_day_domain",u"type",b"type",u"url_domain",b"url_domain",u"value_count",b"value_count"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"annotation",b"annotation",u"bool_domain",b"bool_domain",u"deprecated",b"deprecated",u"distribution_constraints",b"distribution_constraints",u"domain",b"domain",u"domain_info",b"domain_info",u"drift_comparator",b"drift_comparator",u"float_domain",b"float_domain",u"group_presence",b"group_presence",u"image_domain",b"image_domain",u"in_environment",b"in_environment",u"int_domain",b"int_domain",u"lifecycle_stage",b"lifecycle_stage",u"mid_domain",b"mid_domain",u"name",b"name",u"natural_language_domain",b"natural_language_domain",u"not_in_environment",b"not_in_environment",u"presence",b"presence",u"presence_constraints",b"presence_constraints",u"shape",b"shape",u"shape_type",b"shape_type",u"skew_comparator",b"skew_comparator",u"string_domain",b"string_domain",u"struct_domain",b"struct_domain",u"time_domain",b"time_domain",u"time_of_day_domain",b"time_of_day_domain",u"type",b"type",u"url_domain",b"url_domain",u"value_count",b"value_count"]) -> None: ...
    @typing___overload
    def WhichOneof(self, oneof_group: typing_extensions___Literal[u"domain_info",b"domain_info"]) -> typing_extensions___Literal["domain","int_domain","float_domain","string_domain","bool_domain","struct_domain","natural_language_domain","image_domain","mid_domain","url_domain","time_domain","time_of_day_domain"]: ...
    @typing___overload
    def WhichOneof(self, oneof_group: typing_extensions___Literal[u"presence_constraints",b"presence_constraints"]) -> typing_extensions___Literal["presence","group_presence"]: ...
    @typing___overload
    def WhichOneof(self, oneof_group: typing_extensions___Literal[u"shape_type",b"shape_type"]) -> typing_extensions___Literal["shape","value_count"]: ...
type___Feature = Feature

class Annotation(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    tag: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text] = ...
    comment: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text] = ...

    @property
    def extra_metadata(self) -> google___protobuf___internal___containers___RepeatedCompositeFieldContainer[google___protobuf___any_pb2___Any]: ...

    def __init__(self,
        *,
        tag : typing___Optional[typing___Iterable[typing___Text]] = None,
        comment : typing___Optional[typing___Iterable[typing___Text]] = None,
        extra_metadata : typing___Optional[typing___Iterable[google___protobuf___any_pb2___Any]] = None,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"comment",b"comment",u"extra_metadata",b"extra_metadata",u"tag",b"tag"]) -> None: ...
type___Annotation = Annotation

class NumericValueComparator(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    min_fraction_threshold: builtin___float = ...
    max_fraction_threshold: builtin___float = ...

    def __init__(self,
        *,
        min_fraction_threshold : typing___Optional[builtin___float] = None,
        max_fraction_threshold : typing___Optional[builtin___float] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"max_fraction_threshold",b"max_fraction_threshold",u"min_fraction_threshold",b"min_fraction_threshold"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"max_fraction_threshold",b"max_fraction_threshold",u"min_fraction_threshold",b"min_fraction_threshold"]) -> None: ...
type___NumericValueComparator = NumericValueComparator

class DatasetConstraints(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    min_examples_count: builtin___int = ...

    @property
    def num_examples_drift_comparator(self) -> type___NumericValueComparator: ...

    @property
    def num_examples_version_comparator(self) -> type___NumericValueComparator: ...

    def __init__(self,
        *,
        num_examples_drift_comparator : typing___Optional[type___NumericValueComparator] = None,
        num_examples_version_comparator : typing___Optional[type___NumericValueComparator] = None,
        min_examples_count : typing___Optional[builtin___int] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"min_examples_count",b"min_examples_count",u"num_examples_drift_comparator",b"num_examples_drift_comparator",u"num_examples_version_comparator",b"num_examples_version_comparator"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"min_examples_count",b"min_examples_count",u"num_examples_drift_comparator",b"num_examples_drift_comparator",u"num_examples_version_comparator",b"num_examples_version_comparator"]) -> None: ...
type___DatasetConstraints = DatasetConstraints

class FixedShape(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class Dim(google___protobuf___message___Message):
        DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
        size: builtin___int = ...
        name: typing___Text = ...

        def __init__(self,
            *,
            size : typing___Optional[builtin___int] = None,
            name : typing___Optional[typing___Text] = None,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions___Literal[u"name",b"name",u"size",b"size"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"name",b"name",u"size",b"size"]) -> None: ...
    type___Dim = Dim


    @property
    def dim(self) -> google___protobuf___internal___containers___RepeatedCompositeFieldContainer[type___FixedShape.Dim]: ...

    def __init__(self,
        *,
        dim : typing___Optional[typing___Iterable[type___FixedShape.Dim]] = None,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"dim",b"dim"]) -> None: ...
type___FixedShape = FixedShape

class ValueCount(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    min: builtin___int = ...
    max: builtin___int = ...

    def __init__(self,
        *,
        min : typing___Optional[builtin___int] = None,
        max : typing___Optional[builtin___int] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"max",b"max",u"min",b"min"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"max",b"max",u"min",b"min"]) -> None: ...
type___ValueCount = ValueCount

class WeightedFeature(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    name: typing___Text = ...
    lifecycle_stage: type___LifecycleStageValue = ...

    @property
    def feature(self) -> tensorflow_metadata___proto___v0___path_pb2___Path: ...

    @property
    def weight_feature(self) -> tensorflow_metadata___proto___v0___path_pb2___Path: ...

    def __init__(self,
        *,
        name : typing___Optional[typing___Text] = None,
        feature : typing___Optional[tensorflow_metadata___proto___v0___path_pb2___Path] = None,
        weight_feature : typing___Optional[tensorflow_metadata___proto___v0___path_pb2___Path] = None,
        lifecycle_stage : typing___Optional[type___LifecycleStageValue] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"feature",b"feature",u"lifecycle_stage",b"lifecycle_stage",u"name",b"name",u"weight_feature",b"weight_feature"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"feature",b"feature",u"lifecycle_stage",b"lifecycle_stage",u"name",b"name",u"weight_feature",b"weight_feature"]) -> None: ...
type___WeightedFeature = WeightedFeature

class SparseFeature(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class IndexFeature(google___protobuf___message___Message):
        DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
        name: typing___Text = ...

        def __init__(self,
            *,
            name : typing___Optional[typing___Text] = None,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions___Literal[u"name",b"name"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"name",b"name"]) -> None: ...
    type___IndexFeature = IndexFeature

    class ValueFeature(google___protobuf___message___Message):
        DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
        name: typing___Text = ...

        def __init__(self,
            *,
            name : typing___Optional[typing___Text] = None,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions___Literal[u"name",b"name"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"name",b"name"]) -> None: ...
    type___ValueFeature = ValueFeature

    name: typing___Text = ...
    deprecated: builtin___bool = ...
    lifecycle_stage: type___LifecycleStageValue = ...
    is_sorted: builtin___bool = ...
    type: type___FeatureTypeValue = ...

    @property
    def presence(self) -> type___FeaturePresence: ...

    @property
    def dense_shape(self) -> type___FixedShape: ...

    @property
    def index_feature(self) -> google___protobuf___internal___containers___RepeatedCompositeFieldContainer[type___SparseFeature.IndexFeature]: ...

    @property
    def value_feature(self) -> type___SparseFeature.ValueFeature: ...

    def __init__(self,
        *,
        name : typing___Optional[typing___Text] = None,
        deprecated : typing___Optional[builtin___bool] = None,
        lifecycle_stage : typing___Optional[type___LifecycleStageValue] = None,
        presence : typing___Optional[type___FeaturePresence] = None,
        dense_shape : typing___Optional[type___FixedShape] = None,
        index_feature : typing___Optional[typing___Iterable[type___SparseFeature.IndexFeature]] = None,
        is_sorted : typing___Optional[builtin___bool] = None,
        value_feature : typing___Optional[type___SparseFeature.ValueFeature] = None,
        type : typing___Optional[type___FeatureTypeValue] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"dense_shape",b"dense_shape",u"deprecated",b"deprecated",u"is_sorted",b"is_sorted",u"lifecycle_stage",b"lifecycle_stage",u"name",b"name",u"presence",b"presence",u"type",b"type",u"value_feature",b"value_feature"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"dense_shape",b"dense_shape",u"deprecated",b"deprecated",u"index_feature",b"index_feature",u"is_sorted",b"is_sorted",u"lifecycle_stage",b"lifecycle_stage",u"name",b"name",u"presence",b"presence",u"type",b"type",u"value_feature",b"value_feature"]) -> None: ...
type___SparseFeature = SparseFeature

class DistributionConstraints(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    min_domain_mass: builtin___float = ...

    def __init__(self,
        *,
        min_domain_mass : typing___Optional[builtin___float] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"min_domain_mass",b"min_domain_mass"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"min_domain_mass",b"min_domain_mass"]) -> None: ...
type___DistributionConstraints = DistributionConstraints

class IntDomain(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    name: typing___Text = ...
    min: builtin___int = ...
    max: builtin___int = ...
    is_categorical: builtin___bool = ...

    def __init__(self,
        *,
        name : typing___Optional[typing___Text] = None,
        min : typing___Optional[builtin___int] = None,
        max : typing___Optional[builtin___int] = None,
        is_categorical : typing___Optional[builtin___bool] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"is_categorical",b"is_categorical",u"max",b"max",u"min",b"min",u"name",b"name"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"is_categorical",b"is_categorical",u"max",b"max",u"min",b"min",u"name",b"name"]) -> None: ...
type___IntDomain = IntDomain

class FloatDomain(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    name: typing___Text = ...
    min: builtin___float = ...
    max: builtin___float = ...

    def __init__(self,
        *,
        name : typing___Optional[typing___Text] = None,
        min : typing___Optional[builtin___float] = None,
        max : typing___Optional[builtin___float] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"max",b"max",u"min",b"min",u"name",b"name"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"max",b"max",u"min",b"min",u"name",b"name"]) -> None: ...
type___FloatDomain = FloatDomain

class StructDomain(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def feature(self) -> google___protobuf___internal___containers___RepeatedCompositeFieldContainer[type___Feature]: ...

    @property
    def sparse_feature(self) -> google___protobuf___internal___containers___RepeatedCompositeFieldContainer[type___SparseFeature]: ...

    def __init__(self,
        *,
        feature : typing___Optional[typing___Iterable[type___Feature]] = None,
        sparse_feature : typing___Optional[typing___Iterable[type___SparseFeature]] = None,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"feature",b"feature",u"sparse_feature",b"sparse_feature"]) -> None: ...
type___StructDomain = StructDomain

class StringDomain(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    name: typing___Text = ...
    value: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text] = ...

    def __init__(self,
        *,
        name : typing___Optional[typing___Text] = None,
        value : typing___Optional[typing___Iterable[typing___Text]] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"name",b"name"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"name",b"name",u"value",b"value"]) -> None: ...
type___StringDomain = StringDomain

class BoolDomain(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    name: typing___Text = ...
    true_value: typing___Text = ...
    false_value: typing___Text = ...

    def __init__(self,
        *,
        name : typing___Optional[typing___Text] = None,
        true_value : typing___Optional[typing___Text] = None,
        false_value : typing___Optional[typing___Text] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"false_value",b"false_value",u"name",b"name",u"true_value",b"true_value"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"false_value",b"false_value",u"name",b"name",u"true_value",b"true_value"]) -> None: ...
type___BoolDomain = BoolDomain

class NaturalLanguageDomain(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    def __init__(self,
        ) -> None: ...
type___NaturalLanguageDomain = NaturalLanguageDomain

class ImageDomain(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    def __init__(self,
        ) -> None: ...
type___ImageDomain = ImageDomain

class MIDDomain(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    def __init__(self,
        ) -> None: ...
type___MIDDomain = MIDDomain

class URLDomain(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    def __init__(self,
        ) -> None: ...
type___URLDomain = URLDomain

class TimeDomain(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    IntegerTimeFormatValue = typing___NewType('IntegerTimeFormatValue', builtin___int)
    type___IntegerTimeFormatValue = IntegerTimeFormatValue
    IntegerTimeFormat: _IntegerTimeFormat
    class _IntegerTimeFormat(google___protobuf___internal___enum_type_wrapper____EnumTypeWrapper[TimeDomain.IntegerTimeFormatValue]):
        DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
        FORMAT_UNKNOWN = typing___cast(TimeDomain.IntegerTimeFormatValue, 0)
        UNIX_DAYS = typing___cast(TimeDomain.IntegerTimeFormatValue, 5)
        UNIX_SECONDS = typing___cast(TimeDomain.IntegerTimeFormatValue, 1)
        UNIX_MILLISECONDS = typing___cast(TimeDomain.IntegerTimeFormatValue, 2)
        UNIX_MICROSECONDS = typing___cast(TimeDomain.IntegerTimeFormatValue, 3)
        UNIX_NANOSECONDS = typing___cast(TimeDomain.IntegerTimeFormatValue, 4)
    FORMAT_UNKNOWN = typing___cast(TimeDomain.IntegerTimeFormatValue, 0)
    UNIX_DAYS = typing___cast(TimeDomain.IntegerTimeFormatValue, 5)
    UNIX_SECONDS = typing___cast(TimeDomain.IntegerTimeFormatValue, 1)
    UNIX_MILLISECONDS = typing___cast(TimeDomain.IntegerTimeFormatValue, 2)
    UNIX_MICROSECONDS = typing___cast(TimeDomain.IntegerTimeFormatValue, 3)
    UNIX_NANOSECONDS = typing___cast(TimeDomain.IntegerTimeFormatValue, 4)

    string_format: typing___Text = ...
    integer_format: type___TimeDomain.IntegerTimeFormatValue = ...

    def __init__(self,
        *,
        string_format : typing___Optional[typing___Text] = None,
        integer_format : typing___Optional[type___TimeDomain.IntegerTimeFormatValue] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"format",b"format",u"integer_format",b"integer_format",u"string_format",b"string_format"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"format",b"format",u"integer_format",b"integer_format",u"string_format",b"string_format"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions___Literal[u"format",b"format"]) -> typing_extensions___Literal["string_format","integer_format"]: ...
type___TimeDomain = TimeDomain

class TimeOfDayDomain(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    IntegerTimeOfDayFormatValue = typing___NewType('IntegerTimeOfDayFormatValue', builtin___int)
    type___IntegerTimeOfDayFormatValue = IntegerTimeOfDayFormatValue
    IntegerTimeOfDayFormat: _IntegerTimeOfDayFormat
    class _IntegerTimeOfDayFormat(google___protobuf___internal___enum_type_wrapper____EnumTypeWrapper[TimeOfDayDomain.IntegerTimeOfDayFormatValue]):
        DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
        FORMAT_UNKNOWN = typing___cast(TimeOfDayDomain.IntegerTimeOfDayFormatValue, 0)
        PACKED_64_NANOS = typing___cast(TimeOfDayDomain.IntegerTimeOfDayFormatValue, 1)
    FORMAT_UNKNOWN = typing___cast(TimeOfDayDomain.IntegerTimeOfDayFormatValue, 0)
    PACKED_64_NANOS = typing___cast(TimeOfDayDomain.IntegerTimeOfDayFormatValue, 1)

    string_format: typing___Text = ...
    integer_format: type___TimeOfDayDomain.IntegerTimeOfDayFormatValue = ...

    def __init__(self,
        *,
        string_format : typing___Optional[typing___Text] = None,
        integer_format : typing___Optional[type___TimeOfDayDomain.IntegerTimeOfDayFormatValue] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"format",b"format",u"integer_format",b"integer_format",u"string_format",b"string_format"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"format",b"format",u"integer_format",b"integer_format",u"string_format",b"string_format"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions___Literal[u"format",b"format"]) -> typing_extensions___Literal["string_format","integer_format"]: ...
type___TimeOfDayDomain = TimeOfDayDomain

class FeaturePresence(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    min_fraction: builtin___float = ...
    min_count: builtin___int = ...

    def __init__(self,
        *,
        min_fraction : typing___Optional[builtin___float] = None,
        min_count : typing___Optional[builtin___int] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"min_count",b"min_count",u"min_fraction",b"min_fraction"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"min_count",b"min_count",u"min_fraction",b"min_fraction"]) -> None: ...
type___FeaturePresence = FeaturePresence

class FeaturePresenceWithinGroup(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    required: builtin___bool = ...

    def __init__(self,
        *,
        required : typing___Optional[builtin___bool] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"required",b"required"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"required",b"required"]) -> None: ...
type___FeaturePresenceWithinGroup = FeaturePresenceWithinGroup

class InfinityNorm(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    threshold: builtin___float = ...

    def __init__(self,
        *,
        threshold : typing___Optional[builtin___float] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"threshold",b"threshold"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"threshold",b"threshold"]) -> None: ...
type___InfinityNorm = InfinityNorm

class FeatureComparator(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def infinity_norm(self) -> type___InfinityNorm: ...

    def __init__(self,
        *,
        infinity_norm : typing___Optional[type___InfinityNorm] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"infinity_norm",b"infinity_norm"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"infinity_norm",b"infinity_norm"]) -> None: ...
type___FeatureComparator = FeatureComparator

class TensorRepresentation(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class DefaultValue(google___protobuf___message___Message):
        DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
        float_value: builtin___float = ...
        int_value: builtin___int = ...
        bytes_value: builtin___bytes = ...
        uint_value: builtin___int = ...

        def __init__(self,
            *,
            float_value : typing___Optional[builtin___float] = None,
            int_value : typing___Optional[builtin___int] = None,
            bytes_value : typing___Optional[builtin___bytes] = None,
            uint_value : typing___Optional[builtin___int] = None,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions___Literal[u"bytes_value",b"bytes_value",u"float_value",b"float_value",u"int_value",b"int_value",u"kind",b"kind",u"uint_value",b"uint_value"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"bytes_value",b"bytes_value",u"float_value",b"float_value",u"int_value",b"int_value",u"kind",b"kind",u"uint_value",b"uint_value"]) -> None: ...
        def WhichOneof(self, oneof_group: typing_extensions___Literal[u"kind",b"kind"]) -> typing_extensions___Literal["float_value","int_value","bytes_value","uint_value"]: ...
    type___DefaultValue = DefaultValue

    class DenseTensor(google___protobuf___message___Message):
        DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
        column_name: typing___Text = ...

        @property
        def shape(self) -> type___FixedShape: ...

        @property
        def default_value(self) -> type___TensorRepresentation.DefaultValue: ...

        def __init__(self,
            *,
            column_name : typing___Optional[typing___Text] = None,
            shape : typing___Optional[type___FixedShape] = None,
            default_value : typing___Optional[type___TensorRepresentation.DefaultValue] = None,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions___Literal[u"column_name",b"column_name",u"default_value",b"default_value",u"shape",b"shape"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"column_name",b"column_name",u"default_value",b"default_value",u"shape",b"shape"]) -> None: ...
    type___DenseTensor = DenseTensor

    class VarLenSparseTensor(google___protobuf___message___Message):
        DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
        column_name: typing___Text = ...

        def __init__(self,
            *,
            column_name : typing___Optional[typing___Text] = None,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions___Literal[u"column_name",b"column_name"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"column_name",b"column_name"]) -> None: ...
    type___VarLenSparseTensor = VarLenSparseTensor

    class SparseTensor(google___protobuf___message___Message):
        DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
        index_column_names: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text] = ...
        value_column_name: typing___Text = ...

        @property
        def dense_shape(self) -> type___FixedShape: ...

        def __init__(self,
            *,
            dense_shape : typing___Optional[type___FixedShape] = None,
            index_column_names : typing___Optional[typing___Iterable[typing___Text]] = None,
            value_column_name : typing___Optional[typing___Text] = None,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions___Literal[u"dense_shape",b"dense_shape",u"value_column_name",b"value_column_name"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"dense_shape",b"dense_shape",u"index_column_names",b"index_column_names",u"value_column_name",b"value_column_name"]) -> None: ...
    type___SparseTensor = SparseTensor


    @property
    def dense_tensor(self) -> type___TensorRepresentation.DenseTensor: ...

    @property
    def varlen_sparse_tensor(self) -> type___TensorRepresentation.VarLenSparseTensor: ...

    @property
    def sparse_tensor(self) -> type___TensorRepresentation.SparseTensor: ...

    def __init__(self,
        *,
        dense_tensor : typing___Optional[type___TensorRepresentation.DenseTensor] = None,
        varlen_sparse_tensor : typing___Optional[type___TensorRepresentation.VarLenSparseTensor] = None,
        sparse_tensor : typing___Optional[type___TensorRepresentation.SparseTensor] = None,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"dense_tensor",b"dense_tensor",u"kind",b"kind",u"sparse_tensor",b"sparse_tensor",u"varlen_sparse_tensor",b"varlen_sparse_tensor"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"dense_tensor",b"dense_tensor",u"kind",b"kind",u"sparse_tensor",b"sparse_tensor",u"varlen_sparse_tensor",b"varlen_sparse_tensor"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions___Literal[u"kind",b"kind"]) -> typing_extensions___Literal["dense_tensor","varlen_sparse_tensor","sparse_tensor"]: ...
type___TensorRepresentation = TensorRepresentation

class TensorRepresentationGroup(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class TensorRepresentationEntry(google___protobuf___message___Message):
        DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
        key: typing___Text = ...

        @property
        def value(self) -> type___TensorRepresentation: ...

        def __init__(self,
            *,
            key : typing___Optional[typing___Text] = None,
            value : typing___Optional[type___TensorRepresentation] = None,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions___Literal[u"key",b"key",u"value",b"value"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"key",b"key",u"value",b"value"]) -> None: ...
    type___TensorRepresentationEntry = TensorRepresentationEntry


    @property
    def tensor_representation(self) -> google___protobuf___internal___containers___MessageMap[typing___Text, type___TensorRepresentation]: ...

    def __init__(self,
        *,
        tensor_representation : typing___Optional[typing___Mapping[typing___Text, type___TensorRepresentation]] = None,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"tensor_representation",b"tensor_representation"]) -> None: ...
type___TensorRepresentationGroup = TensorRepresentationGroup



import random
import string
import subprocess
import sys
import tempfile
from contextlib import contextmanager
from pathlib import Path
from textwrap import dedent
from typing import List, Tuple

from feast import cli
from feast.feature_store import FeatureStore


def get_example_repo(example_repo_py) -> str:
    parent = Path(__file__).parent
    traversal_limit = 5
    while traversal_limit > 0 and parent.parts[-1] != "tests":
        traversal_limit -= 1
        parent = parent.parent

    if parent.parts[-1] != "tests":
        raise ValueError(f"Unable to find where repo {example_repo_py} is located")

    return (parent / "example_repos" / example_repo_py).read_text()


class CliRunner:
    """
    NB. We can't use test runner helper from click here, since it doesn't start a new Python
    interpreter. And we need a new interpreter for each test since we dynamically import
    modules from the feature repo, and it is hard to clean up that state otherwise.
    """

    def run(self, args: List[str], cwd: Path) -> subprocess.CompletedProcess:
        return subprocess.run([sys.executable, cli.__file__] + args, cwd=cwd)

    def run_with_output(self, args: List[str], cwd: Path) -> Tuple[int, bytes]:
        try:
            return (
                0,
                subprocess.check_output(
                    [sys.executable, cli.__file__] + args,
                    cwd=cwd,
                    stderr=subprocess.STDOUT,
                ),
            )
        except subprocess.CalledProcessError as e:
            return e.returncode, e.output

    @contextmanager
    def local_repo(self, example_repo_py: str, offline_store: str):
        """
        Convenience method to set up all the boilerplate for a local feature repo.
        """
        project_id = "test" + "".join(
            random.choice(string.ascii_lowercase + string.digits) for _ in range(10)
        )

        with tempfile.TemporaryDirectory() as repo_dir_name, tempfile.TemporaryDirectory() as data_dir_name:

            repo_path = Path(repo_dir_name)
            data_path = Path(data_dir_name)

            repo_config = repo_path / "feature_store.yaml"

            repo_config.write_text(
                dedent(
                    f"""
            project: {project_id}
            registry: {data_path / "registry.db"}
            provider: local
            online_store:
                path: {data_path / "online_store.db"}
            offline_store:
                type: {offline_store}
            """
                )
            )

            repo_example = repo_path / "example.py"
            repo_example.write_text(example_repo_py)

            result = self.run(["apply"], cwd=repo_path)
            assert result.returncode == 0

            yield FeatureStore(repo_path=str(repo_path), config=None)

            result = self.run(["teardown"], cwd=repo_path)
            assert result.returncode == 0



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/core/RequestFeatureView.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from feast.protos.feast.core import FeatureView_pb2 as feast_dot_core_dot_FeatureView__pb2
from feast.protos.feast.core import Feature_pb2 as feast_dot_core_dot_Feature__pb2
from feast.protos.feast.core import DataSource_pb2 as feast_dot_core_dot_DataSource__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='feast/core/RequestFeatureView.proto',
  package='feast.core',
  syntax='proto3',
  serialized_options=b'\n\020feast.proto.coreB\027RequestFeatureViewProtoZ3github.com/feast-dev/feast/sdk/go/protos/feast/core',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n#feast/core/RequestFeatureView.proto\x12\nfeast.core\x1a\x1c\x66\x65\x61st/core/FeatureView.proto\x1a\x18\x66\x65\x61st/core/Feature.proto\x1a\x1b\x66\x65\x61st/core/DataSource.proto\"F\n\x12RequestFeatureView\x12\x30\n\x04spec\x18\x01 \x01(\x0b\x32\".feast.core.RequestFeatureViewSpec\"l\n\x16RequestFeatureViewSpec\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07project\x18\x02 \x01(\t\x12\x33\n\x13request_data_source\x18\x03 \x01(\x0b\x32\x16.feast.core.DataSourceB`\n\x10\x66\x65\x61st.proto.coreB\x17RequestFeatureViewProtoZ3github.com/feast-dev/feast/sdk/go/protos/feast/coreb\x06proto3'
  ,
  dependencies=[feast_dot_core_dot_FeatureView__pb2.DESCRIPTOR,feast_dot_core_dot_Feature__pb2.DESCRIPTOR,feast_dot_core_dot_DataSource__pb2.DESCRIPTOR,])




_REQUESTFEATUREVIEW = _descriptor.Descriptor(
  name='RequestFeatureView',
  full_name='feast.core.RequestFeatureView',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='spec', full_name='feast.core.RequestFeatureView.spec', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=136,
  serialized_end=206,
)


_REQUESTFEATUREVIEWSPEC = _descriptor.Descriptor(
  name='RequestFeatureViewSpec',
  full_name='feast.core.RequestFeatureViewSpec',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='feast.core.RequestFeatureViewSpec.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project', full_name='feast.core.RequestFeatureViewSpec.project', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request_data_source', full_name='feast.core.RequestFeatureViewSpec.request_data_source', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=208,
  serialized_end=316,
)

_REQUESTFEATUREVIEW.fields_by_name['spec'].message_type = _REQUESTFEATUREVIEWSPEC
_REQUESTFEATUREVIEWSPEC.fields_by_name['request_data_source'].message_type = feast_dot_core_dot_DataSource__pb2._DATASOURCE
DESCRIPTOR.message_types_by_name['RequestFeatureView'] = _REQUESTFEATUREVIEW
DESCRIPTOR.message_types_by_name['RequestFeatureViewSpec'] = _REQUESTFEATUREVIEWSPEC
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

RequestFeatureView = _reflection.GeneratedProtocolMessageType('RequestFeatureView', (_message.Message,), {
  'DESCRIPTOR' : _REQUESTFEATUREVIEW,
  '__module__' : 'feast.core.RequestFeatureView_pb2'
  # @@protoc_insertion_point(class_scope:feast.core.RequestFeatureView)
  })
_sym_db.RegisterMessage(RequestFeatureView)

RequestFeatureViewSpec = _reflection.GeneratedProtocolMessageType('RequestFeatureViewSpec', (_message.Message,), {
  'DESCRIPTOR' : _REQUESTFEATUREVIEWSPEC,
  '__module__' : 'feast.core.RequestFeatureView_pb2'
  # @@protoc_insertion_point(class_scope:feast.core.RequestFeatureViewSpec)
  })
_sym_db.RegisterMessage(RequestFeatureViewSpec)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import tempfile
from datetime import datetime, timedelta
from pathlib import Path

from tests.utils.cli_utils import CliRunner


def test_cli_chdir() -> None:
    """
    This test simply makes sure that you can run 'feast --chdir COMMAND'
    to switch to a feature repository before running a COMMAND.
    """
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as temp_dir:
        # Make sure the path is absolute by resolving any symlinks
        temp_path = Path(temp_dir).resolve()
        result = runner.run(["init", "my_project"], cwd=temp_path)
        repo_path = temp_path / "my_project"
        assert result.returncode == 0

        result = runner.run(["--chdir", repo_path, "apply"], cwd=temp_path)
        assert result.returncode == 0

        result = runner.run(["--chdir", repo_path, "entities", "list"], cwd=temp_path)
        assert result.returncode == 0

        result = runner.run(
            ["--chdir", repo_path, "feature-views", "list"], cwd=temp_path
        )
        assert result.returncode == 0

        end_date = datetime.utcnow()
        start_date = end_date - timedelta(days=100)
        result = runner.run(
            [
                "--chdir",
                repo_path,
                "materialize",
                start_date.isoformat(),
                end_date.isoformat(),
            ],
            cwd=temp_path,
        )
        assert result.returncode == 0

        result = runner.run(
            ["--chdir", repo_path, "materialize-incremental", end_date.isoformat()],
            cwd=temp_path,
        )
        assert result.returncode == 0

        result = runner.run(["--chdir", repo_path, "registry-dump"], cwd=temp_path)
        assert result.returncode == 0

        result = runner.run(["--chdir", repo_path, "teardown"], cwd=temp_path)
        assert result.returncode == 0



from google.protobuf.duration_pb2 import Duration

from feast import Entity, Feature, FeatureView, FileSource, ValueType

driver_hourly_stats = FileSource(
    path="%PARQUET_PATH%",  # placeholder to be replaced by the test
    event_timestamp_column="event_timestamp",
    created_timestamp_column="created",
)


# The join key here is deliberately different from the parquet file to test the failure path.
driver = Entity(
    name="driver_id",
    value_type=ValueType.INT64,
    description="driver id",
    join_key="driver",
)


driver_hourly_stats_view = FeatureView(
    name="driver_hourly_stats",
    entities=["driver_id"],
    ttl=Duration(seconds=86400 * 1),
    features=[
        Feature(name="conv_rate", dtype=ValueType.FLOAT),
        Feature(name="acc_rate", dtype=ValueType.FLOAT),
        Feature(name="avg_daily_trips", dtype=ValueType.INT64),
    ],
    online=True,
    batch_source=driver_hourly_stats,
    tags={},
)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/serving/TransformationService.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='feast/serving/TransformationService.proto',
  package='feast.serving',
  syntax='proto3',
  serialized_options=b'\n\023feast.proto.servingB\035TransformationServiceAPIProtoZ6github.com/feast-dev/feast/sdk/go/protos/feast/serving',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n)feast/serving/TransformationService.proto\x12\rfeast.serving\"+\n\tValueType\x12\x15\n\x0b\x61rrow_value\x18\x01 \x01(\x0cH\x00\x42\x07\n\x05value\"%\n#GetTransformationServiceInfoRequest\"\x9c\x01\n$GetTransformationServiceInfoResponse\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x36\n\x04type\x18\x02 \x01(\x0e\x32(.feast.serving.TransformationServiceType\x12+\n#transformation_service_type_details\x18\x03 \x01(\t\"\x88\x01\n\x18TransformFeaturesRequest\x12#\n\x1bon_demand_feature_view_name\x18\x01 \x01(\t\x12\x0f\n\x07project\x18\x02 \x01(\t\x12\x36\n\x14transformation_input\x18\x03 \x01(\x0b\x32\x18.feast.serving.ValueType\"T\n\x19TransformFeaturesResponse\x12\x37\n\x15transformation_output\x18\x03 \x01(\x0b\x32\x18.feast.serving.ValueType*\x94\x01\n\x19TransformationServiceType\x12\'\n#TRANSFORMATION_SERVICE_TYPE_INVALID\x10\x00\x12&\n\"TRANSFORMATION_SERVICE_TYPE_PYTHON\x10\x01\x12&\n\"TRANSFORMATION_SERVICE_TYPE_CUSTOM\x10\x64\x32\x89\x02\n\x15TransformationService\x12\x87\x01\n\x1cGetTransformationServiceInfo\x12\x32.feast.serving.GetTransformationServiceInfoRequest\x1a\x33.feast.serving.GetTransformationServiceInfoResponse\x12\x66\n\x11TransformFeatures\x12\'.feast.serving.TransformFeaturesRequest\x1a(.feast.serving.TransformFeaturesResponseBl\n\x13\x66\x65\x61st.proto.servingB\x1dTransformationServiceAPIProtoZ6github.com/feast-dev/feast/sdk/go/protos/feast/servingb\x06proto3'
)

_TRANSFORMATIONSERVICETYPE = _descriptor.EnumDescriptor(
  name='TransformationServiceType',
  full_name='feast.serving.TransformationServiceType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TRANSFORMATION_SERVICE_TYPE_INVALID', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRANSFORMATION_SERVICE_TYPE_PYTHON', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRANSFORMATION_SERVICE_TYPE_CUSTOM', index=2, number=100,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=529,
  serialized_end=677,
)
_sym_db.RegisterEnumDescriptor(_TRANSFORMATIONSERVICETYPE)

TransformationServiceType = enum_type_wrapper.EnumTypeWrapper(_TRANSFORMATIONSERVICETYPE)
TRANSFORMATION_SERVICE_TYPE_INVALID = 0
TRANSFORMATION_SERVICE_TYPE_PYTHON = 1
TRANSFORMATION_SERVICE_TYPE_CUSTOM = 100



_VALUETYPE = _descriptor.Descriptor(
  name='ValueType',
  full_name='feast.serving.ValueType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='arrow_value', full_name='feast.serving.ValueType.arrow_value', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='feast.serving.ValueType.value',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=60,
  serialized_end=103,
)


_GETTRANSFORMATIONSERVICEINFOREQUEST = _descriptor.Descriptor(
  name='GetTransformationServiceInfoRequest',
  full_name='feast.serving.GetTransformationServiceInfoRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=105,
  serialized_end=142,
)


_GETTRANSFORMATIONSERVICEINFORESPONSE = _descriptor.Descriptor(
  name='GetTransformationServiceInfoResponse',
  full_name='feast.serving.GetTransformationServiceInfoResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='feast.serving.GetTransformationServiceInfoResponse.version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='feast.serving.GetTransformationServiceInfoResponse.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='transformation_service_type_details', full_name='feast.serving.GetTransformationServiceInfoResponse.transformation_service_type_details', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=145,
  serialized_end=301,
)


_TRANSFORMFEATURESREQUEST = _descriptor.Descriptor(
  name='TransformFeaturesRequest',
  full_name='feast.serving.TransformFeaturesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='on_demand_feature_view_name', full_name='feast.serving.TransformFeaturesRequest.on_demand_feature_view_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project', full_name='feast.serving.TransformFeaturesRequest.project', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='transformation_input', full_name='feast.serving.TransformFeaturesRequest.transformation_input', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=304,
  serialized_end=440,
)


_TRANSFORMFEATURESRESPONSE = _descriptor.Descriptor(
  name='TransformFeaturesResponse',
  full_name='feast.serving.TransformFeaturesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='transformation_output', full_name='feast.serving.TransformFeaturesResponse.transformation_output', index=0,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=442,
  serialized_end=526,
)

_VALUETYPE.oneofs_by_name['value'].fields.append(
  _VALUETYPE.fields_by_name['arrow_value'])
_VALUETYPE.fields_by_name['arrow_value'].containing_oneof = _VALUETYPE.oneofs_by_name['value']
_GETTRANSFORMATIONSERVICEINFORESPONSE.fields_by_name['type'].enum_type = _TRANSFORMATIONSERVICETYPE
_TRANSFORMFEATURESREQUEST.fields_by_name['transformation_input'].message_type = _VALUETYPE
_TRANSFORMFEATURESRESPONSE.fields_by_name['transformation_output'].message_type = _VALUETYPE
DESCRIPTOR.message_types_by_name['ValueType'] = _VALUETYPE
DESCRIPTOR.message_types_by_name['GetTransformationServiceInfoRequest'] = _GETTRANSFORMATIONSERVICEINFOREQUEST
DESCRIPTOR.message_types_by_name['GetTransformationServiceInfoResponse'] = _GETTRANSFORMATIONSERVICEINFORESPONSE
DESCRIPTOR.message_types_by_name['TransformFeaturesRequest'] = _TRANSFORMFEATURESREQUEST
DESCRIPTOR.message_types_by_name['TransformFeaturesResponse'] = _TRANSFORMFEATURESRESPONSE
DESCRIPTOR.enum_types_by_name['TransformationServiceType'] = _TRANSFORMATIONSERVICETYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ValueType = _reflection.GeneratedProtocolMessageType('ValueType', (_message.Message,), {
  'DESCRIPTOR' : _VALUETYPE,
  '__module__' : 'feast.serving.TransformationService_pb2'
  # @@protoc_insertion_point(class_scope:feast.serving.ValueType)
  })
_sym_db.RegisterMessage(ValueType)

GetTransformationServiceInfoRequest = _reflection.GeneratedProtocolMessageType('GetTransformationServiceInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTRANSFORMATIONSERVICEINFOREQUEST,
  '__module__' : 'feast.serving.TransformationService_pb2'
  # @@protoc_insertion_point(class_scope:feast.serving.GetTransformationServiceInfoRequest)
  })
_sym_db.RegisterMessage(GetTransformationServiceInfoRequest)

GetTransformationServiceInfoResponse = _reflection.GeneratedProtocolMessageType('GetTransformationServiceInfoResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETTRANSFORMATIONSERVICEINFORESPONSE,
  '__module__' : 'feast.serving.TransformationService_pb2'
  # @@protoc_insertion_point(class_scope:feast.serving.GetTransformationServiceInfoResponse)
  })
_sym_db.RegisterMessage(GetTransformationServiceInfoResponse)

TransformFeaturesRequest = _reflection.GeneratedProtocolMessageType('TransformFeaturesRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRANSFORMFEATURESREQUEST,
  '__module__' : 'feast.serving.TransformationService_pb2'
  # @@protoc_insertion_point(class_scope:feast.serving.TransformFeaturesRequest)
  })
_sym_db.RegisterMessage(TransformFeaturesRequest)

TransformFeaturesResponse = _reflection.GeneratedProtocolMessageType('TransformFeaturesResponse', (_message.Message,), {
  'DESCRIPTOR' : _TRANSFORMFEATURESRESPONSE,
  '__module__' : 'feast.serving.TransformationService_pb2'
  # @@protoc_insertion_point(class_scope:feast.serving.TransformFeaturesResponse)
  })
_sym_db.RegisterMessage(TransformFeaturesResponse)


DESCRIPTOR._options = None

_TRANSFORMATIONSERVICE = _descriptor.ServiceDescriptor(
  name='TransformationService',
  full_name='feast.serving.TransformationService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=680,
  serialized_end=945,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetTransformationServiceInfo',
    full_name='feast.serving.TransformationService.GetTransformationServiceInfo',
    index=0,
    containing_service=None,
    input_type=_GETTRANSFORMATIONSERVICEINFOREQUEST,
    output_type=_GETTRANSFORMATIONSERVICEINFORESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='TransformFeatures',
    full_name='feast.serving.TransformationService.TransformFeatures',
    index=1,
    containing_service=None,
    input_type=_TRANSFORMFEATURESREQUEST,
    output_type=_TRANSFORMFEATURESRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_TRANSFORMATIONSERVICE)

DESCRIPTOR.services_by_name['TransformationService'] = _TRANSFORMATIONSERVICE

# @@protoc_insertion_point(module_scope)



from datetime import timedelta

from feast import BigQuerySource, Entity, Feature, FeatureView, ValueType

# Define an entity for the driver. Entities can be thought of as primary keys used to
# retrieve features. Entities are also used to join multiple tables/views during the
# construction of feature vectors
driver = Entity(
    # Name of the entity. Must be unique within a project
    name="driver_id",
    # The join key of an entity describes the storage level field/column on which
    # features can be looked up. The join key is also used to join feature
    # tables/views when building feature vectors
    join_key="driver_id",
    # The storage level type for an entity
    value_type=ValueType.INT64,
)

# Indicates a data source from which feature values can be retrieved. Sources are queried when building training
# datasets or materializing features into an online store.
driver_stats_source = BigQuerySource(
    # The BigQuery table where features can be found
    table_ref="feast-oss.demo_data.driver_hourly_stats_2",
    # The event timestamp is used for point-in-time joins and for ensuring only
    # features within the TTL are returned
    event_timestamp_column="event_timestamp",
    # The (optional) created timestamp is used to ensure there are no duplicate
    # feature rows in the offline store or when building training datasets
    created_timestamp_column="created",
)

# Feature views are a grouping based on how features are stored in either the
# online or offline store.
driver_stats_fv = FeatureView(
    # The unique name of this feature view. Two feature views in a single
    # project cannot have the same name
    name="driver_hourly_stats",
    # The list of entities specifies the keys required for joining or looking
    # up features from this feature view. The reference provided in this field
    # correspond to the name of a defined entity (or entities)
    entities=["driver_id"],
    # The timedelta is the maximum age that each feature value may have
    # relative to its lookup time. For historical features (used in training),
    # TTL is relative to each timestamp provided in the entity dataframe.
    # TTL also allows for eviction of keys from online stores and limits the
    # amount of historical scanning required for historical feature values
    # during retrieval
    ttl=timedelta(weeks=52),
    # The list of features defined below act as a schema to both define features
    # for both materialization of features into a store, and are used as references
    # during retrieval for building a training dataset or serving features
    features=[
        Feature(name="conv_rate", dtype=ValueType.FLOAT),
        Feature(name="acc_rate", dtype=ValueType.FLOAT),
        Feature(name="avg_daily_trips", dtype=ValueType.INT64),
    ],
    # Batch sources are used to find feature values. In the case of this feature
    # view we will query a source table on BigQuery for driver statistics
    # features
    batch_source=driver_stats_source,
    # Tags are user defined key/value pairs that are attached to each
    # feature view
    tags={"team": "driver_performance"},
)



# Copyright 2019 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import Any, Callable, Optional, Tuple

from feast.constants import MAX_WAIT_INTERVAL


def wait_retry_backoff(
    retry_fn: Callable[[], Tuple[Any, bool]],
    timeout_secs: int = 0,
    timeout_msg: Optional[str] = "Timeout while waiting for retry_fn() to return True",
    max_interval_secs: int = int(MAX_WAIT_INTERVAL),
) -> Any:
    """
    Repeatedly try calling given retry_fn until it returns a True boolean success flag.
    Waits with a exponential backoff between retries until timeout when it throws TimeoutError.
    Args:
        retry_fn: Callable that returns a result and a boolean success flag.
        timeout_secs: timeout in seconds to give up retrying and throw TimeoutError,
                        or 0 to retry perpetually.
        timeout_msg: Message to use when throwing TimeoutError.
        max_interval_secs: max wait in seconds to wait between retries.
    Returns:
        Returned Result from retry_fn() if success flag is True.
    """
    wait_secs, elapsed_secs = 1.0, 0.0
    result, is_success = retry_fn()
    wait_begin = time.time()
    while not is_success and (elapsed_secs <= timeout_secs or timeout_secs == 0):
        # back off wait duration exponentially, capped at MAX_WAIT_INTERVAL_SEC
        elapsed_secs = time.time() - wait_begin
        till_timeout_secs = timeout_secs - elapsed_secs
        wait_secs = min(wait_secs * 2, max_interval_secs, till_timeout_secs)
        time.sleep(wait_secs)
        # retry call
        result, is_success = retry_fn()
        elapsed_secs = time.time() - wait_begin

    if not is_success and elapsed_secs > timeout_secs:
        raise TimeoutError(timeout_msg)

    return result



# Copyright 2020 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional

from feast.protos.feast.core.Feature_pb2 import FeatureSpecV2 as FeatureSpecProto
from feast.protos.feast.serving.ServingService_pb2 import (
    FeatureReferenceV2 as FeatureRefProto,
)
from feast.protos.feast.types import Value_pb2 as ValueTypeProto
from feast.value_type import ValueType


class Feature:
    """
    A Feature represents a class of serveable feature.

    Args:
        name: Name of the feature.
        dtype: The type of the feature, such as string or float.
        labels (optional): User-defined metadata in dictionary form.
    """

    def __init__(
        self, name: str, dtype: ValueType, labels: Optional[Dict[str, str]] = None,
    ):
        """Creates a Feature object."""
        self._name = name
        if not isinstance(dtype, ValueType):
            raise ValueError("dtype is not a valid ValueType")
        if dtype is ValueType.UNKNOWN:
            raise ValueError(f"dtype cannot be {dtype}")
        self._dtype = dtype
        if labels is None:
            self._labels = dict()
        else:
            self._labels = labels

    def __eq__(self, other):
        if (
            self.name != other.name
            or self.dtype != other.dtype
            or self.labels != other.labels
        ):
            return False
        return True

    def __lt__(self, other):
        return self.name < other.name

    def __repr__(self):
        # return string representation of the reference
        return f"{self.name}-{self.dtype}"

    def __str__(self):
        # readable string of the reference
        return f"Feature<{self.__repr__()}>"

    @property
    def name(self):
        """
        Gets the name of this feature.
        """
        return self._name

    @property
    def dtype(self) -> ValueType:
        """
        Gets the data type of this feature.
        """
        return self._dtype

    @property
    def labels(self) -> Dict[str, str]:
        """
        Gets the labels of this feature.
        """
        return self._labels

    def to_proto(self) -> FeatureSpecProto:
        """
        Converts Feature object to its Protocol Buffer representation.

        Returns:
            A FeatureSpecProto protobuf.
        """
        value_type = ValueTypeProto.ValueType.Enum.Value(self.dtype.name)

        return FeatureSpecProto(
            name=self.name, value_type=value_type, labels=self.labels,
        )

    @classmethod
    def from_proto(cls, feature_proto: FeatureSpecProto):
        """
        Args:
            feature_proto: FeatureSpecV2 protobuf object

        Returns:
            Feature object
        """
        feature = cls(
            name=feature_proto.name,
            dtype=ValueType(feature_proto.value_type),
            labels=dict(feature_proto.labels),
        )

        return feature


class FeatureRef:
    """ Feature Reference represents a reference to a specific feature. """

    def __init__(self, name: str, feature_table: str):
        self.proto = FeatureRefProto(name=name, feature_table=feature_table)

    @classmethod
    def from_proto(cls, proto: FeatureRefProto):
        """
        Construct a feature reference from the given FeatureReference proto

        Args:
            proto: Protobuf FeatureReference to construct from
        Returns:
            FeatureRef that refers to the given feature
        """
        return cls(name=proto.name, feature_table=proto.feature_table)

    @classmethod
    def from_str(cls, feature_ref_str: str):
        """
        Parse the given string feature reference into FeatureRef model
        String feature reference should be in the format feature_table:feature.
        Where "feature_table" and "name" are the feature_table name and feature name
        respectively.

        Args:
            feature_ref_str: String representation of the feature reference
        Returns:
            FeatureRef that refers to the given feature
        """
        proto = FeatureRefProto()

        # parse feature table name if specified
        if ":" in feature_ref_str:
            proto.feature_table, proto.name = feature_ref_str.split(":")
        else:
            raise ValueError(
                f"Unsupported feature reference: {feature_ref_str} - Feature reference string should be in the form [featuretable_name:featurename]"
            )

        return cls.from_proto(proto)

    def to_proto(self) -> FeatureRefProto:
        """
        Convert and return this feature table reference to protobuf.

        Returns:
            Protobuf respresentation of this feature table reference.
        """

        return self.proto

    def __repr__(self):
        # return string representation of the reference
        ref_str = self.proto.feature_table + ":" + self.proto.name
        return ref_str

    def __str__(self):
        # readable string of the reference
        return f"FeatureRef<{self.__repr__()}>"


def _build_feature_references(feature_ref_strs: List[str]) -> List[FeatureRefProto]:
    """
    Builds a list of FeatureReference protos from a list of FeatureReference strings

    Args:
        feature_ref_strs: List of string feature references
    Returns:
        A list of FeatureReference protos parsed from args.
    """

    feature_refs = [FeatureRef.from_str(ref_str) for ref_str in feature_ref_strs]
    feature_ref_protos = [ref.to_proto() for ref in feature_refs]

    return feature_ref_protos



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/core/FeatureService.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from feast.protos.feast.core import FeatureViewProjection_pb2 as feast_dot_core_dot_FeatureViewProjection__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='feast/core/FeatureService.proto',
  package='feast.core',
  syntax='proto3',
  serialized_options=b'\n\020feast.proto.coreB\023FeatureServiceProtoZ3github.com/feast-dev/feast/sdk/go/protos/feast/core',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1f\x66\x65\x61st/core/FeatureService.proto\x12\nfeast.core\x1a\x1fgoogle/protobuf/timestamp.proto\x1a&feast/core/FeatureViewProjection.proto\"l\n\x0e\x46\x65\x61tureService\x12,\n\x04spec\x18\x01 \x01(\x0b\x32\x1e.feast.core.FeatureServiceSpec\x12,\n\x04meta\x18\x02 \x01(\x0b\x32\x1e.feast.core.FeatureServiceMeta\"\xe2\x01\n\x12\x46\x65\x61tureServiceSpec\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07project\x18\x02 \x01(\t\x12\x33\n\x08\x66\x65\x61tures\x18\x03 \x03(\x0b\x32!.feast.core.FeatureViewProjection\x12\x36\n\x04tags\x18\x04 \x03(\x0b\x32(.feast.core.FeatureServiceSpec.TagsEntry\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x87\x01\n\x12\x46\x65\x61tureServiceMeta\x12\x35\n\x11\x63reated_timestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12:\n\x16last_updated_timestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\\\n\x10\x66\x65\x61st.proto.coreB\x13\x46\x65\x61tureServiceProtoZ3github.com/feast-dev/feast/sdk/go/protos/feast/coreb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,feast_dot_core_dot_FeatureViewProjection__pb2.DESCRIPTOR,])




_FEATURESERVICE = _descriptor.Descriptor(
  name='FeatureService',
  full_name='feast.core.FeatureService',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='spec', full_name='feast.core.FeatureService.spec', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='meta', full_name='feast.core.FeatureService.meta', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=120,
  serialized_end=228,
)


_FEATURESERVICESPEC_TAGSENTRY = _descriptor.Descriptor(
  name='TagsEntry',
  full_name='feast.core.FeatureServiceSpec.TagsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='feast.core.FeatureServiceSpec.TagsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='feast.core.FeatureServiceSpec.TagsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=414,
  serialized_end=457,
)

_FEATURESERVICESPEC = _descriptor.Descriptor(
  name='FeatureServiceSpec',
  full_name='feast.core.FeatureServiceSpec',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='feast.core.FeatureServiceSpec.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project', full_name='feast.core.FeatureServiceSpec.project', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='features', full_name='feast.core.FeatureServiceSpec.features', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tags', full_name='feast.core.FeatureServiceSpec.tags', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='feast.core.FeatureServiceSpec.description', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_FEATURESERVICESPEC_TAGSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=231,
  serialized_end=457,
)


_FEATURESERVICEMETA = _descriptor.Descriptor(
  name='FeatureServiceMeta',
  full_name='feast.core.FeatureServiceMeta',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='created_timestamp', full_name='feast.core.FeatureServiceMeta.created_timestamp', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_updated_timestamp', full_name='feast.core.FeatureServiceMeta.last_updated_timestamp', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=460,
  serialized_end=595,
)

_FEATURESERVICE.fields_by_name['spec'].message_type = _FEATURESERVICESPEC
_FEATURESERVICE.fields_by_name['meta'].message_type = _FEATURESERVICEMETA
_FEATURESERVICESPEC_TAGSENTRY.containing_type = _FEATURESERVICESPEC
_FEATURESERVICESPEC.fields_by_name['features'].message_type = feast_dot_core_dot_FeatureViewProjection__pb2._FEATUREVIEWPROJECTION
_FEATURESERVICESPEC.fields_by_name['tags'].message_type = _FEATURESERVICESPEC_TAGSENTRY
_FEATURESERVICEMETA.fields_by_name['created_timestamp'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_FEATURESERVICEMETA.fields_by_name['last_updated_timestamp'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
DESCRIPTOR.message_types_by_name['FeatureService'] = _FEATURESERVICE
DESCRIPTOR.message_types_by_name['FeatureServiceSpec'] = _FEATURESERVICESPEC
DESCRIPTOR.message_types_by_name['FeatureServiceMeta'] = _FEATURESERVICEMETA
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

FeatureService = _reflection.GeneratedProtocolMessageType('FeatureService', (_message.Message,), {
  'DESCRIPTOR' : _FEATURESERVICE,
  '__module__' : 'feast.core.FeatureService_pb2'
  # @@protoc_insertion_point(class_scope:feast.core.FeatureService)
  })
_sym_db.RegisterMessage(FeatureService)

FeatureServiceSpec = _reflection.GeneratedProtocolMessageType('FeatureServiceSpec', (_message.Message,), {

  'TagsEntry' : _reflection.GeneratedProtocolMessageType('TagsEntry', (_message.Message,), {
    'DESCRIPTOR' : _FEATURESERVICESPEC_TAGSENTRY,
    '__module__' : 'feast.core.FeatureService_pb2'
    # @@protoc_insertion_point(class_scope:feast.core.FeatureServiceSpec.TagsEntry)
    })
  ,
  'DESCRIPTOR' : _FEATURESERVICESPEC,
  '__module__' : 'feast.core.FeatureService_pb2'
  # @@protoc_insertion_point(class_scope:feast.core.FeatureServiceSpec)
  })
_sym_db.RegisterMessage(FeatureServiceSpec)
_sym_db.RegisterMessage(FeatureServiceSpec.TagsEntry)

FeatureServiceMeta = _reflection.GeneratedProtocolMessageType('FeatureServiceMeta', (_message.Message,), {
  'DESCRIPTOR' : _FEATURESERVICEMETA,
  '__module__' : 'feast.core.FeatureService_pb2'
  # @@protoc_insertion_point(class_scope:feast.core.FeatureServiceMeta)
  })
_sym_db.RegisterMessage(FeatureServiceMeta)


DESCRIPTOR._options = None
_FEATURESERVICESPEC_TAGSENTRY._options = None
# @@protoc_insertion_point(module_scope)



from datetime import datetime
from typing import Dict, List, Optional, Union

from google.protobuf.json_format import MessageToJson

from feast.base_feature_view import BaseFeatureView
from feast.feature_table import FeatureTable
from feast.feature_view import FeatureView
from feast.feature_view_projection import FeatureViewProjection
from feast.on_demand_feature_view import OnDemandFeatureView
from feast.protos.feast.core.FeatureService_pb2 import (
    FeatureService as FeatureServiceProto,
)
from feast.protos.feast.core.FeatureService_pb2 import (
    FeatureServiceMeta,
    FeatureServiceSpec,
)
from feast.usage import log_exceptions


class FeatureService:
    """
    A feature service is a logical grouping of features for retrieval (training or serving).
    The features grouped by a feature service may come from any number of feature views.

    Args:
        name: Unique name of the feature service.
        features: A list of Features that are grouped as part of this FeatureService.
            The list may contain Feature Views, Feature Tables, or a subset of either.
        tags (optional): A dictionary of key-value pairs used for organizing Feature
            Services.
    """

    name: str
    feature_view_projections: List[FeatureViewProjection]
    tags: Dict[str, str]
    description: Optional[str] = None
    created_timestamp: Optional[datetime] = None
    last_updated_timestamp: Optional[datetime] = None

    @log_exceptions
    def __init__(
        self,
        name: str,
        features: List[Union[FeatureTable, FeatureView, OnDemandFeatureView]],
        tags: Optional[Dict[str, str]] = None,
        description: Optional[str] = None,
    ):
        """
        Creates a FeatureService object.

        Raises:
            ValueError: If one of the specified features is not a valid type.
        """
        self.name = name
        self.feature_view_projections = []

        for feature_grouping in features:
            if isinstance(feature_grouping, FeatureTable):
                self.feature_view_projections.append(
                    FeatureViewProjection.from_definition(feature_grouping)
                )
            elif isinstance(feature_grouping, BaseFeatureView):
                self.feature_view_projections.append(feature_grouping.projection)
            else:
                raise ValueError(
                    "The FeatureService {fs_name} has been provided with an invalid type"
                    f'{type(feature_grouping)} as part of the "features" argument.)'
                )

        self.tags = tags or {}
        self.description = description
        self.created_timestamp = None
        self.last_updated_timestamp = None

    def __repr__(self):
        items = (f"{k} = {v}" for k, v in self.__dict__.items())
        return f"<{self.__class__.__name__}({', '.join(items)})>"

    def __str__(self):
        return str(MessageToJson(self.to_proto()))

    def __hash__(self):
        return hash((id(self), self.name))

    def __eq__(self, other):
        if not isinstance(other, FeatureService):
            raise TypeError(
                "Comparisons should only involve FeatureService class objects."
            )
        if self.tags != other.tags or self.name != other.name:
            return False

        if sorted(self.feature_view_projections) != sorted(
            other.feature_view_projections
        ):
            return False

        return True

    @staticmethod
    def from_proto(feature_service_proto: FeatureServiceProto):
        """
        Converts a FeatureServiceProto to a FeatureService object.

        Args:
            feature_service_proto: A protobuf representation of a FeatureService.
        """
        fs = FeatureService(
            name=feature_service_proto.spec.name,
            features=[],
            tags=dict(feature_service_proto.spec.tags),
            description=(
                feature_service_proto.spec.description
                if feature_service_proto.spec.description != ""
                else None
            ),
        )
        fs.feature_view_projections.extend(
            [
                FeatureViewProjection.from_proto(projection)
                for projection in feature_service_proto.spec.features
            ]
        )

        if feature_service_proto.meta.HasField("created_timestamp"):
            fs.created_timestamp = (
                feature_service_proto.meta.created_timestamp.ToDatetime()
            )
        if feature_service_proto.meta.HasField("last_updated_timestamp"):
            fs.last_updated_timestamp = (
                feature_service_proto.meta.last_updated_timestamp.ToDatetime()
            )

        return fs

    def to_proto(self) -> FeatureServiceProto:
        """
        Converts a FeatureService to its protobuf representation.

        Returns:
            A FeatureServiceProto protobuf.
        """
        meta = FeatureServiceMeta()
        if self.created_timestamp:
            meta.created_timestamp.FromDatetime(self.created_timestamp)

        spec = FeatureServiceSpec(
            name=self.name,
            features=[
                projection.to_proto() for projection in self.feature_view_projections
            ],
        )

        if self.tags:
            spec.tags.update(self.tags)
        if self.description:
            spec.description = self.description

        feature_service_proto = FeatureServiceProto(spec=spec, meta=meta)
        return feature_service_proto

    def validate(self):
        pass



# Copyright 2021 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import multiprocessing
from datetime import datetime, timedelta
from sys import platform
from typing import List

import pandas as pd
import pytest
from _pytest.nodes import Item

from tests.data.data_creator import create_dataset
from tests.integration.feature_repos.integration_test_repo_config import (
    IntegrationTestRepoConfig,
)
from tests.integration.feature_repos.repo_configuration import (
    FULL_REPO_CONFIGS,
    REDIS_CONFIG,
    Environment,
    construct_test_environment,
    construct_universal_data_sources,
    construct_universal_datasets,
    construct_universal_entities,
)


def pytest_configure(config):
    if platform in ["darwin", "windows"]:
        multiprocessing.set_start_method("spawn")
    else:
        multiprocessing.set_start_method("fork")
    config.addinivalue_line(
        "markers", "integration: mark test that has external dependencies"
    )
    config.addinivalue_line("markers", "benchmark: mark benchmarking tests")
    config.addinivalue_line(
        "markers", "universal: mark tests that use the universal feature repo"
    )


def pytest_addoption(parser):
    parser.addoption(
        "--integration",
        action="store_true",
        default=False,
        help="Run tests with external dependencies",
    )
    parser.addoption(
        "--benchmark", action="store_true", default=False, help="Run benchmark tests",
    )
    parser.addoption(
        "--universal", action="store_true", default=False, help="Run universal tests",
    )


def pytest_collection_modifyitems(config, items: List[Item]):
    should_run_integration = config.getoption("--integration") is True
    should_run_benchmark = config.getoption("--benchmark") is True
    should_run_universal = config.getoption("--universal") is True

    integration_tests = [t for t in items if "integration" in t.keywords]
    if not should_run_integration:
        for t in integration_tests:
            items.remove(t)
    else:
        items.clear()
        for t in integration_tests:
            items.append(t)

    benchmark_tests = [t for t in items if "benchmark" in t.keywords]
    if not should_run_benchmark:
        for t in benchmark_tests:
            items.remove(t)
    else:
        items.clear()
        for t in benchmark_tests:
            items.append(t)

    universal_tests = [t for t in items if "universal" in t.keywords]
    if should_run_universal:
        items.clear()
        for t in universal_tests:
            items.append(t)


@pytest.fixture
def simple_dataset_1() -> pd.DataFrame:
    now = datetime.utcnow()
    ts = pd.Timestamp(now).round("ms")
    data = {
        "id": [1, 2, 1, 3, 3],
        "float_col": [0.1, 0.2, 0.3, 4, 5],
        "int64_col": [1, 2, 3, 4, 5],
        "string_col": ["a", "b", "c", "d", "e"],
        "ts_1": [
            ts,
            ts - timedelta(hours=4),
            ts - timedelta(hours=3),
            ts - timedelta(hours=2),
            ts - timedelta(hours=1),
        ],
    }
    return pd.DataFrame.from_dict(data)


@pytest.fixture
def simple_dataset_2() -> pd.DataFrame:
    now = datetime.utcnow()
    ts = pd.Timestamp(now).round("ms")
    data = {
        "id": ["a", "b", "c", "d", "e"],
        "float_col": [0.1, 0.2, 0.3, 4, 5],
        "int64_col": [1, 2, 3, 4, 5],
        "string_col": ["a", "b", "c", "d", "e"],
        "ts_1": [
            ts,
            ts - timedelta(hours=4),
            ts - timedelta(hours=3),
            ts - timedelta(hours=2),
            ts - timedelta(hours=1),
        ],
    }
    return pd.DataFrame.from_dict(data)


@pytest.fixture(
    params=FULL_REPO_CONFIGS, scope="session", ids=[str(c) for c in FULL_REPO_CONFIGS]
)
def environment(request):
    with construct_test_environment(request.param) as e:
        yield e


@pytest.fixture()
def local_redis_environment():
    with construct_test_environment(
        IntegrationTestRepoConfig(online_store=REDIS_CONFIG)
    ) as e:
        yield e


@pytest.fixture(scope="session")
def universal_data_sources(environment):
    entities = construct_universal_entities()
    datasets = construct_universal_datasets(
        entities, environment.start_date, environment.end_date
    )
    datasources = construct_universal_data_sources(
        datasets, environment.data_source_creator
    )

    yield entities, datasets, datasources

    environment.data_source_creator.teardown()


@pytest.fixture(scope="session")
def e2e_data_sources(environment: Environment):
    df = create_dataset()
    data_source = environment.data_source_creator.create_data_source(
        df, environment.feature_store.project, field_mapping={"ts_1": "ts"},
    )

    yield df, data_source

    environment.data_source_creator.teardown()



# Copyright 2021 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
from datetime import datetime
from enum import Enum
from typing import (
    Any,
    ByteString,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from google.protobuf.timestamp_pb2 import Timestamp
from pydantic import StrictStr
from pydantic.typing import Literal

from feast import Entity, FeatureTable, FeatureView, RepoConfig, utils
from feast.infra.online_stores.helpers import _mmh3, _redis_key, _redis_key_prefix
from feast.infra.online_stores.online_store import OnlineStore
from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from feast.repo_config import FeastConfigBaseModel
from feast.usage import log_exceptions_and_usage, tracing_span

try:
    from redis import Redis
    from rediscluster import RedisCluster
except ImportError as e:
    from feast.errors import FeastExtrasDependencyImportError

    raise FeastExtrasDependencyImportError("redis", str(e))

logger = logging.getLogger(__name__)


class RedisType(str, Enum):
    redis = "redis"
    redis_cluster = "redis_cluster"


class RedisOnlineStoreConfig(FeastConfigBaseModel):
    """Online store config for Redis store"""

    type: Literal["redis"] = "redis"
    """Online store type selector"""

    redis_type: RedisType = RedisType.redis
    """Redis type: redis or redis_cluster"""

    connection_string: StrictStr = "localhost:6379"
    """Connection string containing the host, port, and configuration parameters for Redis
     format: host:port,parameter1,parameter2 eg. redis:6379,db=0 """


class RedisOnlineStore(OnlineStore):
    _client: Optional[Union[Redis, RedisCluster]] = None

    def delete_table_values(
        self, config: RepoConfig, table: Union[FeatureTable, FeatureView]
    ):
        client = self._get_client(config.online_store)
        deleted_count = 0
        pipeline = client.pipeline()
        prefix = _redis_key_prefix(table.entities)

        for _k in client.scan_iter(
            b"".join([prefix, b"*", config.project.encode("utf8")])
        ):
            pipeline.delete(_k)
            deleted_count += 1
        pipeline.execute()

        logger.debug(f"Deleted {deleted_count} keys for {table.name}")

    @log_exceptions_and_usage(online_store="redis")
    def update(
        self,
        config: RepoConfig,
        tables_to_delete: Sequence[Union[FeatureTable, FeatureView]],
        tables_to_keep: Sequence[Union[FeatureTable, FeatureView]],
        entities_to_delete: Sequence[Entity],
        entities_to_keep: Sequence[Entity],
        partial: bool,
    ):
        """
        We delete the keys in redis for tables/views being removed.
        """
        for table in tables_to_delete:
            self.delete_table_values(config, table)

    def teardown(
        self,
        config: RepoConfig,
        tables: Sequence[Union[FeatureTable, FeatureView]],
        entities: Sequence[Entity],
    ):
        """
        We delete the keys in redis for tables/views being removed.
        """
        for table in tables:
            self.delete_table_values(config, table)

    @staticmethod
    def _parse_connection_string(connection_string: str):
        """
        Reads Redis connections string using format
        for RedisCluster:
            redis1:6379,redis2:6379,decode_responses=true,skip_full_coverage_check=true,ssl=true,password=...
        for Redis:
            redis_master:6379,db=0,ssl=true,password=...
        """
        startup_nodes = [
            dict(zip(["host", "port"], c.split(":")))
            for c in connection_string.split(",")
            if "=" not in c
        ]
        params = {}
        for c in connection_string.split(","):
            if "=" in c:
                kv = c.split("=", 1)
                try:
                    kv[1] = json.loads(kv[1])
                except json.JSONDecodeError:
                    ...

                it = iter(kv)
                params.update(dict(zip(it, it)))

        return startup_nodes, params

    def _get_client(self, online_store_config: RedisOnlineStoreConfig):
        """
        Creates the Redis client RedisCluster or Redis depending on configuration
        """
        if not self._client:
            startup_nodes, kwargs = self._parse_connection_string(
                online_store_config.connection_string
            )
            if online_store_config.redis_type == RedisType.redis_cluster:
                kwargs["startup_nodes"] = startup_nodes
                self._client = RedisCluster(**kwargs)
            else:
                kwargs["host"] = startup_nodes[0]["host"]
                kwargs["port"] = startup_nodes[0]["port"]
                self._client = Redis(**kwargs)
        return self._client

    @log_exceptions_and_usage(online_store="redis")
    def online_write_batch(
        self,
        config: RepoConfig,
        table: Union[FeatureTable, FeatureView],
        data: List[
            Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
        ],
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        online_store_config = config.online_store
        assert isinstance(online_store_config, RedisOnlineStoreConfig)

        client = self._get_client(online_store_config)
        project = config.project

        feature_view = table.name
        ts_key = f"_ts:{feature_view}"
        keys = []
        # redis pipelining optimization: send multiple commands to redis server without waiting for every reply
        with client.pipeline() as pipe:
            # check if a previous record under the key bin exists
            # TODO: investigate if check and set is a better approach rather than pulling all entity ts and then setting
            # it may be significantly slower but avoids potential (rare) race conditions
            for entity_key, _, _, _ in data:
                redis_key_bin = _redis_key(project, entity_key)
                keys.append(redis_key_bin)
                pipe.hmget(redis_key_bin, ts_key)
            prev_event_timestamps = pipe.execute()
            # flattening the list of lists. `hmget` does the lookup assuming a list of keys in the key bin
            prev_event_timestamps = [i[0] for i in prev_event_timestamps]

            for redis_key_bin, prev_event_time, (_, values, timestamp, _) in zip(
                keys, prev_event_timestamps, data
            ):
                event_time_seconds = int(utils.make_tzaware(timestamp).timestamp())

                # ignore if event_timestamp is before the event features that are currently in the feature store
                if prev_event_time:
                    prev_ts = Timestamp()
                    prev_ts.ParseFromString(prev_event_time)
                    if prev_ts.seconds and event_time_seconds <= prev_ts.seconds:
                        # TODO: somehow signal that it's not overwriting the current record?
                        if progress:
                            progress(1)
                        continue

                ts = Timestamp()
                ts.seconds = event_time_seconds
                entity_hset = dict()
                entity_hset[ts_key] = ts.SerializeToString()

                for feature_name, val in values.items():
                    f_key = _mmh3(f"{feature_view}:{feature_name}")
                    entity_hset[f_key] = val.SerializeToString()

                pipe.hset(redis_key_bin, mapping=entity_hset)
                # TODO: support expiring the entity / features in Redis
                # otherwise entity features remain in redis until cleaned up in separate process
                # client.expire redis_key_bin based a ttl setting
            results = pipe.execute()
            if progress:
                progress(len(results))

    @log_exceptions_and_usage(online_store="redis")
    def online_read(
        self,
        config: RepoConfig,
        table: Union[FeatureTable, FeatureView],
        entity_keys: List[EntityKeyProto],
        requested_features: Optional[List[str]] = None,
    ) -> List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]]:
        online_store_config = config.online_store
        assert isinstance(online_store_config, RedisOnlineStoreConfig)

        client = self._get_client(online_store_config)
        feature_view = table.name
        project = config.project

        result: List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]] = []

        if not requested_features:
            requested_features = [f.name for f in table.features]

        hset_keys = [_mmh3(f"{feature_view}:{k}") for k in requested_features]

        ts_key = f"_ts:{feature_view}"
        hset_keys.append(ts_key)
        requested_features.append(ts_key)

        keys = []
        for entity_key in entity_keys:
            redis_key_bin = _redis_key(project, entity_key)
            keys.append(redis_key_bin)
        with client.pipeline() as pipe:
            for redis_key_bin in keys:
                pipe.hmget(redis_key_bin, hset_keys)
            with tracing_span(name="remote_call"):
                redis_values = pipe.execute()
        for values in redis_values:
            features = self._get_features_for_entity(
                values, feature_view, requested_features
            )
            result.append(features)
        return result

    def _get_features_for_entity(
        self,
        values: List[ByteString],
        feature_view: str,
        requested_features: List[str],
    ) -> Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]:
        res_val = dict(zip(requested_features, values))

        res_ts = Timestamp()
        ts_val = res_val.pop(f"_ts:{feature_view}")
        if ts_val:
            res_ts.ParseFromString(ts_val)

        res = {}
        for feature_name, val_bin in res_val.items():
            val = ValueProto()
            if val_bin:
                val.ParseFromString(val_bin)
            res[feature_name] = val

        if not res:
            return None, None
        else:
            timestamp = datetime.fromtimestamp(res_ts.seconds)
            return timestamp, res
